# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# mypy: disable-error-code="misc,arg-type,type-arg,valid-type,assignment,return-value"
"""torch.ops.aten operators under the `special` module.

- No inplace operators.
- All functions should not have the script() decorator. This is because
    we want to delay the compilation of the function.
"""

from __future__ import annotations

import math
from typing import Optional, Sequence

from onnxscript.function_libs.torch_lib.registration import torch_op
from onnxscript.function_libs.torch_lib.tensor_typing import TFloat
from onnxscript.onnx_opset import opset18 as op
from onnxscript.onnx_types import TensorType

_MATH_PI = math.pi


def aten_special_airy_ai(x: TensorType) -> TensorType:
    """special_airy_ai(Tensor x) -> Tensor"""

    raise NotImplementedError()


def aten_special_bessel_j0(self: TensorType) -> TensorType:
    """special_bessel_j0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_bessel_j1(self: TensorType) -> TensorType:
    """special_bessel_j1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_bessel_y0(self: TensorType) -> TensorType:
    """special_bessel_y0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_bessel_y1(self: TensorType) -> TensorType:
    """special_bessel_y1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_chebyshev_polynomial_t(x: TensorType, n: TensorType) -> TensorType:
    """special_chebyshev_polynomial_t(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_chebyshev_polynomial_u(x: TensorType, n: TensorType) -> TensorType:
    """special_chebyshev_polynomial_u(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_chebyshev_polynomial_v(x: TensorType, n: TensorType) -> TensorType:
    """special_chebyshev_polynomial_v(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_chebyshev_polynomial_w(x: TensorType, n: TensorType) -> TensorType:
    """special_chebyshev_polynomial_w(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_digamma(self: TensorType) -> TensorType:
    """special_digamma(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_entr(self: TensorType) -> TensorType:
    """special_entr(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op(("aten::erf", "aten::special_erf"))
def aten_special_erf(self: TFloat) -> TFloat:
    """erf(Tensor self) -> Tensor"""

    return op.Erf(self)


@torch_op(("aten::erfc", "aten::special_erfc"))
def aten_special_erfc(self: TFloat) -> TFloat:
    """erfc(Tensor self) -> Tensor"""

    return op.Sub(1, op.Erf(self))


@torch_op("aten::special_erfcx")
def aten_special_erfcx(self: TFloat) -> TFloat:
    """special_erfcx(Tensor self) -> Tensor"""

    return op.Mul(op.Exp(op.Pow(self, 2)), op.Sub(1, op.Erf(self)))


def aten_special_erfinv(self: TensorType) -> TensorType:
    """special_erfinv(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_exp2(self: TensorType) -> TensorType:
    """special_exp2(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_expit(self: TensorType) -> TensorType:
    """special_expit(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op(("aten::expm1", "aten::special_expm1"))
def aten_special_expm1(self: TFloat) -> TFloat:
    """special_expm1(Tensor self) -> Tensor"""

    return op.Sub(op.Exp(self), 1)


def aten_special_gammainc(self: TensorType, other: TensorType) -> TensorType:
    """special_gammainc(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def aten_special_gammaincc(self: TensorType, other: TensorType) -> TensorType:
    """special_gammaincc(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def aten_special_gammaln(self: TensorType) -> TensorType:
    """special_gammaln(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_hermite_polynomial_h(x: TensorType, n: TensorType) -> TensorType:
    """special_hermite_polynomial_h(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_hermite_polynomial_he(x: TensorType, n: TensorType) -> TensorType:
    """special_hermite_polynomial_he(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_i0(self: TensorType) -> TensorType:
    """special_i0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_i0e(self: TensorType) -> TensorType:
    """special_i0e(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_i1(self: TensorType) -> TensorType:
    """special_i1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_i1e(self: TensorType) -> TensorType:
    """special_i1e(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_laguerre_polynomial_l(x: TensorType, n: TensorType) -> TensorType:
    """special_laguerre_polynomial_l(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_legendre_polynomial_p(x: TensorType, n: TensorType) -> TensorType:
    """special_legendre_polynomial_p(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_log1p(self: TensorType) -> TensorType:
    """special_log1p(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_log_ndtr(self: TensorType) -> TensorType:
    """special_log_ndtr(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op(("aten::log_softmax.int", "aten::special_log_softmax"), trace_only=True)
def aten_special_log_softmax(self: TFloat, dim: int, dtype: int = -1) -> TFloat:
    """special_log_softmax(Tensor self, int dim, *, ScalarType? dtype=None) -> Tensor"""

    self_is_scalar = len(self.shape) == 0
    if self_is_scalar:
        self = op.Unsqueeze(self, [0])
    result = op.LogSoftmax(self, axis=dim)
    if dtype != -1:
        result = op.Cast(result, to=dtype)
    if self_is_scalar:  # squeeze to scalar due to input is scalar
        result = op.Squeeze(result)
    return result


def aten_special_logit(self: TensorType, eps: Optional[float] = None) -> TensorType:
    """special_logit(Tensor self, float? eps=None) -> Tensor"""
    # TODO: alias of core.aten_logit
    raise NotImplementedError()


def aten_special_logsumexp(
    self: TensorType, dim: Sequence[int], keepdim: bool = False
) -> TensorType:
    """special_logsumexp(Tensor self, int[1] dim, bool keepdim=False) -> Tensor"""

    raise NotImplementedError()


def aten_special_modified_bessel_i0(self: TensorType) -> TensorType:
    """special_modified_bessel_i0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_modified_bessel_i1(self: TensorType) -> TensorType:
    """special_modified_bessel_i1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_modified_bessel_k0(self: TensorType) -> TensorType:
    """special_modified_bessel_k0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_modified_bessel_k1(self: TensorType) -> TensorType:
    """special_modified_bessel_k1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_multigammaln(self: TensorType, p: int) -> TensorType:
    """special_multigammaln(Tensor self, int p) -> Tensor"""

    raise NotImplementedError()


def aten_special_ndtr(self: TensorType) -> TensorType:
    """special_ndtr(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_ndtri(self: TensorType) -> TensorType:
    """special_ndtri(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_polygamma(n: int, self: TensorType) -> TensorType:
    """special_polygamma(int n, Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_psi(self: TensorType) -> TensorType:
    """special_psi(Tensor self) -> Tensor"""

    raise NotImplementedError()


def aten_special_round(self: TensorType, decimals: int = 0) -> TensorType:
    """special_round(Tensor self, *, int decimals=0) -> Tensor"""

    raise NotImplementedError()


def aten_special_scaled_modified_bessel_k0(x: TensorType) -> TensorType:
    """special_scaled_modified_bessel_k0(Tensor x) -> Tensor"""

    raise NotImplementedError()


def aten_special_scaled_modified_bessel_k1(x: TensorType) -> TensorType:
    """special_scaled_modified_bessel_k1(Tensor x) -> Tensor"""

    raise NotImplementedError()


def aten_special_shifted_chebyshev_polynomial_t(x: TensorType, n: TensorType) -> TensorType:
    """special_shifted_chebyshev_polynomial_t(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_shifted_chebyshev_polynomial_u(x: TensorType, n: TensorType) -> TensorType:
    """special_shifted_chebyshev_polynomial_u(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_shifted_chebyshev_polynomial_v(x: TensorType, n: TensorType) -> TensorType:
    """special_shifted_chebyshev_polynomial_v(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


def aten_special_shifted_chebyshev_polynomial_w(x: TensorType, n: TensorType) -> TensorType:
    """special_shifted_chebyshev_polynomial_w(Tensor x, Tensor n) -> Tensor"""

    raise NotImplementedError()


@torch_op(("aten::special_sinc", "aten::sinc"))
def aten_special_sinc(self: TFloat) -> TFloat:
    """special_sinc(Tensor self) -> Tensor"""

    # This computes the normalized sinc, where the input is multiplied by pi.
    # https://pytorch.org/docs/stable/special.html#torch.special.sinc
    pi_self = self * _MATH_PI

    return op.Where(self == 0.0, op.CastLike(1, self), op.Sin(pi_self) / pi_self)


def aten_special_spherical_bessel_j0(x: TensorType) -> TensorType:
    """special_spherical_bessel_j0(Tensor x) -> Tensor"""

    raise NotImplementedError()


def aten_special_xlog1py(self: TensorType, other: TensorType) -> TensorType:
    """special_xlog1py(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


@torch_op(("aten::xlogy.Tensor", "aten::xlogy.Scalar_Self", "aten::xlogy.Scalar_Other"))
def aten_special_xlogy(self: TFloat, other: TFloat) -> TFloat:
    """special_xlogy(Tensor self, Tensor other) -> Tensor"""

    # https://pytorch.org/docs/stable/special.html#torch.special.xlogy
    # out := {
    #     NaN if other == NaN
    #     0 if self == 0
    #     self * log(other) otherwise
    # }

    nans = op.IsNaN(other)
    zeros = op.Equal(self, 0)
    xlogy = op.Mul(self, op.Log(other))
    xlogy_with_nans = op.Where(nans, other, xlogy)
    return op.Where(zeros, self, xlogy_with_nans)


def aten_special_zeta(self: TensorType, other: TensorType) -> TensorType:
    """special_zeta(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# pylint: disable=import-outside-toplevel
from __future__ import annotations

import multiprocessing
import os


def get_memory_rss(pid: int) -> int:
    """
    Returns the physical memory used by a process.

    Args:
        pid: Process id, current one is `os.getpid()`.

    Returns:
         Physical memory.

    It relies on the module *psutil*.
    """
    import psutil

    process = psutil.Process(pid)
    mem = process.memory_info().rss
    return mem


class Monitor:
    def __init__(self):
        self.max_peak: float = 0
        self.average: float = 0
        self.n_measures: int = 0
        self.begin: float = 0
        self.end: float = 0

    def to_dict(self, unit: int = 1) -> dict[str, float]:
        funit = float(unit)
        return dict(
            peak=self.max_peak / funit,
            mean=self.average * 1.0 / self.n_measures / funit,
            n=self.n_measures / funit,
            begin=self.begin / funit,
            end=self.end / funit,
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(peak={self.max_peak}, "
            f"average={self.average}, n={self.n_measures})"
        )

    def update(self, mem: float):
        if self.n_measures == 0:
            self.begin = mem
        self.max_peak = max(mem, self.max_peak)
        self.average += mem
        self.end = mem
        self.n_measures += 1

    def send(self, conn):
        conn.send(self.max_peak)
        conn.send(self.average)
        conn.send(self.n_measures)
        conn.send(self.begin)
        conn.send(self.end)

    @classmethod
    def recv(cls, conn) -> Monitor:
        m = cls()
        m.max_peak = conn.recv()
        m.average = conn.recv()
        m.n_measures = conn.recv()
        m.begin = conn.recv()
        m.end = conn.recv()
        return m


def _process_memory_spy(conn):
    # Sends the value it started.
    conn.send(-2)

    # process id to spy on
    pid = conn.recv()

    # delay between two measures
    timeout = conn.recv()

    # do CUDA
    cuda = conn.recv()

    import psutil

    process = psutil.Process(pid)

    if cuda:
        from pynvml import (  # type: ignore[import-not-found]
            nvmlDeviceGetCount,
            nvmlDeviceGetHandleByIndex,
            nvmlDeviceGetMemoryInfo,
            nvmlInit,
            nvmlShutdown,
        )

        nvmlInit()
        n_gpus = nvmlDeviceGetCount()
        handles = [nvmlDeviceGetHandleByIndex(i) for i in range(n_gpus)]

        def gpu_used():
            return [nvmlDeviceGetMemoryInfo(h).used for h in handles]

        gpus = [Monitor() for i in range(n_gpus)]
    else:
        gpus = []

    cpu = Monitor()

    conn.send(-2)

    # loop
    while True:
        mem = process.memory_info().rss
        cpu.update(mem)
        if cuda:
            for r, g in zip(gpu_used(), gpus):
                g.update(r)
        if conn.poll(timeout=timeout):
            code = conn.recv()
            if code == -3:
                break

    # final iteration
    end = process.memory_info().rss
    cpu.update(end)
    if cuda:
        for r, g in zip(gpu_used(), gpus):
            g.update(r)

    # send
    cpu.send(conn)
    conn.send(len(gpus))
    for g in gpus:
        g.send(conn)
    if cuda:
        nvmlShutdown()
    conn.close()


class MemorySpy:
    """
    Information about the spy. It class method `start`.
    Method `stop` can be called to end the measure.

    Args:
        pid: process id  of the process to spy on
        delay: spy on every delay seconds
        cuda: enable cuda monitoring
    """

    def __init__(self, pid: int, delay: float = 0.01, cuda: bool = False):
        self.pid = pid
        self.delay = delay
        self.cuda = cuda
        self.start()

    def start(self) -> MemorySpy:
        """Starts another process and tells it to spy."""
        self.parent_conn, self.child_conn = multiprocessing.Pipe()
        self.child_process = multiprocessing.Process(
            target=_process_memory_spy, args=(self.child_conn,)
        )
        self.child_process.start()
        data = self.parent_conn.recv()
        if data != -2:
            raise RuntimeError(f"The child processing is supposed to send -2 not {data}.")
        self.parent_conn.send(self.pid)
        self.parent_conn.send(self.delay)
        self.parent_conn.send(1 if self.cuda else 0)
        data = self.parent_conn.recv()
        if data != -2:
            raise RuntimeError(
                f"The child processing is supposed to send -2 again not {data}."
            )
        return self

    def stop(self) -> dict[str, list[Monitor]]:
        """Stops spying on."""
        self.parent_conn.send(-3)

        cpu = [Monitor.recv(self.parent_conn)]

        n_gpus = self.parent_conn.recv()
        gpus = []
        for _ in range(n_gpus):
            gpus.append(Monitor.recv(self.parent_conn))

        self.parent_conn.close()
        self.child_process.join()
        res = dict(cpu=cpu)
        if self.cuda:
            res["gpus"] = gpus
        return res


def start_spying_on(
    pid: int | None = None, delay: float = 0.01, cuda: bool = False
) -> MemorySpy:
    """Starts the memory spy. The function starts another
    process spying on the one sent as an argument.

    Example::

    .. code-block:: python

        from onnxscript.tools.memory_peak import start_spying_on, flatten

        p = start_spying_on()
        # ...
        # code to measure
        # ...
        stat = p.stop()
        print(stat)
        print(flatten(stat))

    Args:
        pid: process id to spy or the the current one.
        delay: delay between two measures.
        cuda: True or False to get memory for cuda devices
    """
    if pid is None:
        pid = os.getpid()
    return MemorySpy(pid, delay, cuda)


def flatten(ps: dict[str, list[Monitor]], prefix: str = "") -> dict[str, float]:
    """Flattens a dictionary produced by :meth:`MemorySpy.stop`."""
    obs = ps["cpu"][0].to_dict(unit=2**20)
    if "gpus" in ps:
        for i, g in enumerate(ps["gpus"]):
            for k, v in g.to_dict(unit=2**20).items():
                obs[f"gpu{i}_{k}"] = v
    if prefix:
        obs = {f"{prefix}{k}": v for k, v in obs.items()}
    return obs



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import unittest

import onnx
import onnx_ir as ir
from packaging import version

import onnxscript
import onnxscript.optimizer
import onnxscript.rewriter.testing
from onnxscript import FLOAT, script
from onnxscript.rewriter.rules.fusion._gqa import fuse_gqa

op = onnxscript.values.Opset("", 23)

H = [8]  # Number of attention heads
Hkv = [4]  # Number of key/value heads (H should be divisible by Hkv)
D = [64]  # Head size
G = [2]  # Number of groups


@script(ir_version=10)
def _gqa_script(
    query_BHSD: FLOAT[2, 8, 4, 64],  # B=2, H=8, S=4, D=64
    key_BHkvSD: FLOAT[2, 4, 4, 64],  # B=2, Hkv=4, S=4, D=64
    value_BHkvSD: FLOAT[2, 4, 4, 64],  # B=2, Hkv=4, S=4, D=64
    past_key_BHkvPD: FLOAT[2, 4, 8, 64],  # B=2, Hkv=4, P=8, D=64
    past_value_BHkvPD: FLOAT[2, 4, 8, 64],  # B=2, Hkv=4, P=8, D=64
) -> FLOAT[2, 8, 4, 64]:
    """Basic GQA pattern that should be fused into an Attention op."""

    # Concatenate past_key cache and current key
    present_key_BHkvStD = op.Concat(past_key_BHkvPD, key_BHkvSD, axis=-2)  # [B, Hkv, S+P, D]

    # Unsqueeze to add group dimension
    present_key_BHkv1StD = op.Unsqueeze(present_key_BHkvStD, 2)  # [B, Hkv, 1, S+P, D]

    # Calculate shapes dynamically
    B = op.Shape(query_BHSD, start=0, end=1)  # [B]
    T = op.Shape(present_key_BHkvStD, start=2, end=3)  # [S+P]

    # Create expand shape [B, Hkv, G, S+P, D]
    expand_shape = op.Concat(B, Hkv, G, T, D, axis=0)
    present_key_BHkvGStD = op.Expand(present_key_BHkv1StD, expand_shape)  # [B, Hkv, G, S+P, D]

    # Create reshape shape [B, H, S+P, D]
    reshape_shape = op.Concat(B, H, T, D, axis=0)
    present_key_BHStD = op.Reshape(present_key_BHkvGStD, reshape_shape)  # [B, H, S+P, D]

    # Same for value
    present_value_BHkvStD = op.Concat(
        past_value_BHkvPD, value_BHkvSD, axis=-2
    )  # [B, Hkv, S+P, D]
    present_value_BHkv1StD = op.Unsqueeze(present_value_BHkvStD, 2)  # [B, Hkv, 1, S+P, D]
    present_value_BHkvGStD = op.Expand(
        present_value_BHkv1StD, expand_shape
    )  # [B, Hkv, G, S+P, D]
    present_value_BHStD = op.Reshape(present_value_BHkvGStD, reshape_shape)  # [B, H, S+P, D]

    # Attention computation
    attention_BHSDh = op.Attention(
        query_BHSD,
        present_key_BHStD,
        present_value_BHStD,
    )

    return attention_BHSDh


class GQAFusionTest(unittest.TestCase):
    def test_basic_gqa_fusion(self):
        """Test basic GQA fusion pattern."""
        model_proto = _gqa_script.to_model_proto()

        # Apply GQA fusion
        model = ir.serde.deserialize_model(model_proto)
        onnxscript.optimizer.optimize(model)
        count = fuse_gqa(model)
        self.assertGreater(count, 0, "GQA fusion should have occurred")

        # We can't yet test numerical equivalence because of a bug in the op spec/implementation.
        onnx_ver = version.parse(onnx.__version__)
        if onnx_ver >= version.parse("1.19.1") and not (
            onnx_ver.is_prerelease or onnx_ver.is_devrelease
        ):
            # Only official releases >= 1.19.1
            onnxscript.optimizer.remove_unused_nodes(model)
            rewritten_model_proto = ir.serde.serialize_model(model)
            onnxscript.rewriter.testing.assert_numerically_equal(
                model_proto, rewritten_model_proto, use_reference=True
            )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import numpy as np
import onnx
import parameterized

from onnxscript import ir
from onnxscript.rewriter import testing
from onnxscript.rewriter.rules.common import _fuse_batchnorm


class FuseBatchnormTest(unittest.TestCase):
    def _create_batchnorm_params(self, size: int):
        return [
            onnx.numpy_helper.from_array(
                np.random.randn(size).astype(np.float32), name="gamma"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(size).astype(np.float32), name="beta"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(size).astype(np.float32), name="input_mean"
            ),
            onnx.numpy_helper.from_array(
                np.abs(np.random.randn(size)).astype(np.float32), name="input_var"
            ),
        ]

    @parameterized.parameterized.expand(
        [
            ("bias_false_group1", False, 1),
            ("bias_true_group1", True, 1),
            ("bias_false_group4", False, 4),
            ("bias_true_group4", True, 4),
        ]
    )
    def test_fuse_batchnorm_convtranspose(self, _: str, convtranspose_bias: bool, group: int):
        # ConvTranspose weight: [in_channels, out_channels/group, kH, kW]
        out_channels = 64 * group
        convtranspose_inputs = "X, W"
        parameters = (
            f"float[32, 64, 3, 3] W, "
            f"float[{out_channels}] gamma, "
            f"float[{out_channels}] beta, "
            f"float[{out_channels}] input_mean, "
            f"float[{out_channels}] input_var"
        )
        if convtranspose_bias:
            parameters += f", float[{out_channels}] B"
            convtranspose_inputs += ", B"

        model_proto = onnx.parser.parse_model(f"""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32, 14, 16] X) => (float [N, ?, ?, ?] Y)
            <{parameters}>
            {{
                X1 = ConvTranspose<group={group}>({convtranspose_inputs})
                Y = BatchNormalization(X1, gamma, beta, input_mean, input_var)
            }}
        """)
        # Add initializers
        initializers = [
            onnx.numpy_helper.from_array(
                np.random.randn(32, 64, 3, 3).astype(np.float32), name="W"
            ),
            *self._create_batchnorm_params(size=out_channels),
        ]
        if convtranspose_bias:
            initializers.append(
                onnx.numpy_helper.from_array(
                    np.random.randn(out_channels).astype(np.float32), name="B"
                )
            )
        model_proto.graph.initializer.extend(initializers)

        onnx.checker.check_model(model_proto, True)
        model = ir.serde.deserialize_model(model_proto)

        # Apply rule
        count = _fuse_batchnorm.rules.apply_to_model(model)

        # Check that BatchNorm was fused
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 1)

        # Check inference
        testing.assert_numerically_equal(
            model_proto, model, (np.random.rand(1, 32, 14, 16).astype(np.float32),)
        )

        output_model_proto = ir.serde.serialize_model(model)
        onnx.checker.check_model(output_model_proto, True)

    @parameterized.parameterized.expand(
        [
            ("bias_false_group1", False, 1),
            ("bias_true_group1", True, 1),
            ("bias_false_group2", False, 2),
            ("bias_true_group2", True, 2),
        ]
    )
    def test_fuse_batchnorm_conv(self, _: str, conv_bias: bool, group: int):
        # Conv weight: [out_channels, in_channels/group, kH, kW]
        in_channels_per_group = 32 // group
        conv_inputs = "X, W"
        parameters = (
            f"float[64, {in_channels_per_group}, 3, 3] W, "
            "float[64] gamma, "
            "float[64] beta, "
            "float[64] input_mean, "
            "float[64] input_var"
        )
        if conv_bias:
            parameters += ", float[64] B"
            conv_inputs += ", B"

        model_proto = onnx.parser.parse_model(f"""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32, 14, 16] X) => (float [N, ?, ?, ?] Y)
            <{parameters}>
            {{
                X1 = Conv<group={group}>({conv_inputs})
                Y = BatchNormalization(X1, gamma, beta, input_mean, input_var)
            }}
        """)
        # Add initializers
        initializers = [
            onnx.numpy_helper.from_array(
                np.random.randn(64, in_channels_per_group, 3, 3).astype(np.float32), name="W"
            ),
            *self._create_batchnorm_params(size=64),
        ]
        if conv_bias:
            initializers.append(
                onnx.numpy_helper.from_array(np.random.randn(64).astype(np.float32), name="B")
            )
        model_proto.graph.initializer.extend(initializers)

        onnx.checker.check_model(model_proto, True)
        model = ir.serde.deserialize_model(model_proto)

        # Apply rule
        count = _fuse_batchnorm.rules.apply_to_model(model)

        # Check that BatchNorm was fused
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 1)

        # Check inference
        testing.assert_numerically_equal(
            model_proto, model, (np.random.rand(1, 32, 14, 16).astype(np.float32),)
        )

        output_model_proto = ir.serde.serialize_model(model)
        onnx.checker.check_model(output_model_proto, True)

    @parameterized.parameterized.expand(
        [
            ("bias_false_transB_0", False, 0),
            ("bias_true_transB_0", True, 0),
            ("bias_false_transB_1", False, 1),
            ("bias_true_transB_1", True, 1),
        ]
    )
    def test_fuse_batchnorm_gemm(self, _: str, gemm_bias: bool, transB: int):
        gemm_inputs = "X, W"
        parameters = (
            f"float{'[64, 32]' if transB else '[32, 64]'} W, "
            "float[64] gamma, "
            "float[64] beta, "
            "float[64] input_mean, "
            "float[64] input_var"
        )

        if gemm_bias:
            parameters += ", float[64] B"
            gemm_inputs += ", B"

        model_proto = onnx.parser.parse_model(f"""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32] X) => (float [N, ?] Y)
            <{parameters}>
            {{
                X1 = Gemm<transB={transB}>({gemm_inputs})
                Y = BatchNormalization(X1, gamma, beta, input_mean, input_var)
            }}
        """)
        weights = np.random.randn(32, 64).astype(np.float32)
        if transB:
            weights = weights.T

        # Add initializers
        initializers = [
            onnx.numpy_helper.from_array(weights, name="W"),
            *self._create_batchnorm_params(size=64),
        ]
        if gemm_bias:
            initializers.append(
                onnx.numpy_helper.from_array(np.random.randn(64).astype(np.float32), name="B")
            )
        model_proto.graph.initializer.extend(initializers)

        onnx.checker.check_model(model_proto, True)
        model = ir.serde.deserialize_model(model_proto)

        # Apply rule
        count = _fuse_batchnorm.rules.apply_to_model(model)

        # Check that BatchNorm was fused
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 1)

        # Check inference
        testing.assert_numerically_equal(
            model_proto, model, (np.random.rand(1, 32).astype(np.float32),)
        )

        output_model_proto = ir.serde.serialize_model(model)
        onnx.checker.check_model(output_model_proto, True)

    def test_fuse_batchnorm_convtranspose_grouped_invalid_skipped(self):
        """Fusion is skipped when in_channels is not divisible by group (semantically invalid model)."""
        # in_channels=32 is not divisible by group=3, the ONNX checker won't catch this.
        model_proto = onnx.parser.parse_model("""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32, 14, 14] X) => (float[N, ?, ?, ?] Y)
            <float[32, 64, 3, 3] W,
             float[192] gamma, float[192] beta, float[192] input_mean, float[192] input_var>
            {
                X1 = ConvTranspose<group=3>(X, W)
                Y = BatchNormalization(X1, gamma, beta, input_mean, input_var)
            }
        """)
        initializers = [
            onnx.numpy_helper.from_array(
                np.random.randn(32, 64, 3, 3).astype(np.float32), name="W"
            ),
            *self._create_batchnorm_params(size=192),
        ]
        model_proto.graph.initializer.extend(initializers)
        model = ir.serde.deserialize_model(model_proto)
        count = _fuse_batchnorm.rules.apply_to_model(model)

        # Fusion must be skipped, applying it would crash on the invalid dimensions.
        self.assertEqual(count, 0)

    def test_fuse_batchnorm_non_initializers(self):
        model_proto = onnx.parser.parse_model("""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32, 14, 16] X, float[64, 32, 3, 3] W, float[64] B,
                        float[64] gamma, float[64] beta, float[64] input_var,
                        float[64] input_mean) => (float [N, ?, ?, ?] Y)
            {
                X1 = Conv(X, W, B)
                Y = BatchNormalization(X1, gamma, beta, input_mean, input_var)
            }
        """)
        onnx.checker.check_model(model_proto, True)
        model = ir.serde.deserialize_model(model_proto)
        count = _fuse_batchnorm.rules.apply_to_model(model)

        # No changes were applied
        self.assertEqual(count, 0)

    def test_fuse_batchnorm_graph_inputs(self):
        model_proto = onnx.parser.parse_model("""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32, 14, 16] X, float[64, 32, 3, 3] W) => (float [N, ?, ?, ?] Y)
            {
                X1 = Conv(X, W)
                Y = BatchNormalization(X1, gamma, beta, input_mean, input_var)
            }
        """)
        initializers = [
            onnx.numpy_helper.from_array(
                np.random.randn(64, 32, 3, 3).astype(np.float32), name="W"
            ),
            *self._create_batchnorm_params(size=64),
        ]
        model_proto.graph.initializer.extend(initializers)
        onnx.checker.check_model(model_proto, True)

        model = ir.serde.deserialize_model(model_proto)
        count = _fuse_batchnorm.rules.apply_to_model(model)

        # No changes were applied as W is a graph input
        self.assertEqual(count, 0)

    def test_fuse_batchnorm_does_not_collide_names_with_same_parent_node(self):
        model_proto = onnx.parser.parse_model("""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32, 14, 16] X) => (float [N, ?, ?, ?] Y1, float [N, ?, ?, ?] Y2)
            {
                X1 = MaxPool<kernel_shape=[3,3]>(X)
                X2 = Conv(X1, W1)
                Y1 = BatchNormalization(X2, gamma_64, beta_64, input_mean_64, input_var_64)
                X3 = Conv(X1, W2)
                Y2 = BatchNormalization(X3, gamma_256, beta_256, input_mean_256, input_var_256)
            }
        """)
        initializers = [
            onnx.numpy_helper.from_array(
                np.random.randn(64, 32, 3, 3).astype(np.float32), name="W1"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(64).astype(np.float32), name="gamma_64"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(64).astype(np.float32), name="beta_64"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(64).astype(np.float32), name="input_mean_64"
            ),
            onnx.numpy_helper.from_array(
                np.abs(np.random.randn(64)).astype(np.float32), name="input_var_64"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(256, 32, 3, 3).astype(np.float32), name="W2"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(256).astype(np.float32), name="gamma_256"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(256).astype(np.float32), name="beta_256"
            ),
            onnx.numpy_helper.from_array(
                np.random.randn(256).astype(np.float32), name="input_mean_256"
            ),
            onnx.numpy_helper.from_array(
                np.abs(np.random.randn(256)).astype(np.float32), name="input_var_256"
            ),
        ]
        model_proto.graph.initializer.extend(initializers)
        onnx.checker.check_model(model_proto, True)
        model = ir.serde.deserialize_model(model_proto)
        count = _fuse_batchnorm.rules.apply_to_model(model)

        # Applied twice, once for each BatchNorm
        self.assertEqual(count, 2)
        # it should have different bias names for the two fused Conv nodes
        conv_nodes = [node for node in model.graph if node.op_type == "Conv"]
        self.assertEqual(len(conv_nodes), 2)
        bias_names_1 = conv_nodes[0].inputs[2].name
        bias_names_2 = conv_nodes[1].inputs[2].name
        self.assertNotEqual(bias_names_1, bias_names_2)

    def test_fuse_batchnorm_skips_shared_weight_initializers(self):
        """Test that BatchNorm fusion is skipped when Conv nodes share weight initializers.

        Regression test for https://github.com/microsoft/onnxscript/issues/2382.
        When two Conv+BatchNorm pairs share the same weight initializer, fusing the
        first pair would overwrite the shared initializer, leaving the second Conv
        node with an invalid (unregistered) weight reference.
        """
        model_proto = onnx.parser.parse_model("""
            < ir_version: 7, opset_import: ["" : 17] >
            test_model (float[N, 32, 14, 16] X1, float[N, 32, 14, 16] X2)
                => (float [N, ?, ?, ?] Y)
            {
                C1 = Conv(X1, W, B)
                BN1 = BatchNormalization(C1, gamma, beta, input_mean, input_var)
                C2 = Conv(X2, W, B)
                BN2 = BatchNormalization(C2, gamma, beta, input_mean, input_var)
                Y = Add(BN1, BN2)
            }
        """)
        initializers = [
            onnx.numpy_helper.from_array(
                np.random.randn(16, 32, 3, 3).astype(np.float32), name="W"
            ),
            onnx.numpy_helper.from_array(np.random.randn(16).astype(np.float32), name="B"),
            *self._create_batchnorm_params(size=16),
        ]
        model_proto.graph.initializer.extend(initializers)
        onnx.checker.check_model(model_proto, True)
        model = ir.serde.deserialize_model(model_proto)

        count = _fuse_batchnorm.rules.apply_to_model(model)

        # No fusion should be applied because the weight initializer is shared
        self.assertEqual(count, 0)

        # The model should still be valid after the (non-)optimization
        output_model_proto = ir.serde.serialize_model(model)
        onnx.checker.check_model(output_model_proto, True)

        # Check inference produces correct results
        testing.assert_numerically_equal(
            model_proto,
            model,
            (
                np.random.rand(1, 32, 14, 16).astype(np.float32),
                np.random.rand(1, 32, 14, 16).astype(np.float32),
            ),
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Rewrite rules to eliminate redundant ScatterND operations.

This module contains two rewrite rules:

1. ScatterAllDynamic: Identifies ScatterND(data, indices, updates) that can be replaced by Identity(updates)
   when the indices are computed dynamically using Range operations but represent a complete update
   of an entire axis. This is generated by the translation of `x[:, ...] = y` in PyTorch.

2. ScatterAllStatic: Identifies ScatterND(data, indices, updates) that can be replaced by Identity(updates)
   when the indices are statically known constants in the form [[0], [1], ..., [n-1]] covering
   the entire first dimension of the data tensor.

Both rules detect when the scatter-update ends up being an assignment of a new value to the entire tensor.
"""

from __future__ import annotations

import onnx_ir as ir

import onnxscript.rewriter
from onnxscript.rewriter import _ir_utils
from onnxscript.rewriter._rewrite_rule import RewriteRuleClassBase, RewriteRuleSet


class ScatterAllDynamic(RewriteRuleClassBase):
    def __init__(self):
        super().__init__(remove_nodes=False)

    def pattern(self, op, data, axis, transposed_data, updates):
        # Construct update-indices spanning an entire axis:
        shape = op.Shape(data, start=0)
        dim = op.Gather(shape, axis, axis=0)
        full_range = op.Range(0, dim, 1)
        full_range_2d = op.Unsqueeze(full_range, [-1])
        # The update is applied to the data transposed to bring the updated axis to the front:
        return op.ScatterND(transposed_data, full_range_2d, updates, reduction="none")

    def check(self, context, data, axis, transposed_data, **_):
        # Check that updated-indices represent the full range of the first dimension of the transposed data.
        # That is: check that the data.shape[axis] matches transposed_data.shape[0].
        result = onnxscript.rewriter.MatchResult()
        axis_value = _ir_utils.get_singleton_value(axis)
        if not isinstance(axis_value, int):
            return result.fail("Axis value must be a constant integer.", axis)
        shape: ir.Shape | None = data.shape
        if shape is None:
            return result.fail("Data shape is not statically known.", data)
        updated_dim_value = shape[axis_value]
        transposed_data_shape: ir.Shape | None = transposed_data.shape
        if transposed_data_shape is None:
            return result.fail(
                "Transposed data shape is not statically known.", transposed_data
            )
        actual_dim_value = transposed_data_shape[0]
        if not _ir_utils.same_dim(updated_dim_value, actual_dim_value):
            # The first dimension of the transposed data does not match the updated dimension,
            # so we cannot apply this rule.
            return result.fail(
                "The first dimension of the transposed data does not match the updated dimension.",
                [data, transposed_data],
            )
        return True

    def rewrite(self, op, updates, **_):
        return op.Identity(updates)


class ScatterAllStatic(RewriteRuleClassBase):
    """Rewrite rule for eliminating redundant ScatterND with statically known indices.

    This handles the case where indices are constant values in the form [[0], [1], ..., [n-1]]
    that update the entire first dimension of the data tensor.
    """

    def pattern(self, op, data, indices, updates):
        """Pattern to match ScatterND with static indices."""
        return op.ScatterND(data, indices, updates)

    def check(self, context, data, indices, updates, **_):
        """Check if the ScatterND is redundant due to static indices covering entire tensor."""
        # To validate data can be replaced directly by updates, we need to check the following:
        # 1. they have the same shape
        result = onnxscript.rewriter.MatchResult()
        if data.shape is None:
            return result.fail("The value 'data' shape is not statically known.", data)
        if updates.shape is None:
            return result.fail("The value 'updates' shape is not statically known.", updates)
        if not _ir_utils.same_shape(data.shape, updates.shape):
            return result.fail(
                "The shape of 'data' and 'updates' are different.", [data, updates]
            )

        # 2. the indices is referring to the whole data, which is from 0 to data.shape[0]
        if indices.const_value is None:
            return result.fail("The value 'indices' is not statically known.", indices)
        expected_indices = [[i] for i in range(data.shape[0])]
        actual_indices = indices.const_value.numpy().tolist()
        if actual_indices != expected_indices:
            return result.fail("The 'indices' is not referring to the whole data.", indices)

        return True

    def rewrite(self, op, updates, **_):
        """Replace ScatterND with Identity since updates covers entire tensor."""
        return op.Identity(updates)


no_op_dynamic_scatter_nd_rule = ScatterAllDynamic.rule()
no_op_static_scatter_nd_rule = ScatterAllStatic.rule()

rules = RewriteRuleSet([no_op_dynamic_scatter_nd_rule, no_op_static_scatter_nd_rule])



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, Tuple, TypeVar

from onnx import GraphProto
from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset7 import Opset7
from onnxscript.onnx_types import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset8(Opset7):
    def __new__(cls):
        return Opset.__new__(cls, "", 8)

    T_Expand = TypeVar(
        "T_Expand",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Expand(self, input: T_Expand, shape: INT64) -> T_Expand:
        r"""[🌐 Expand(8)](https://onnx.ai/onnx/operators/onnx__Expand.html#expand-8 "Online Documentation")


        Broadcast the input tensor following the given shape and the broadcast rule.
        The broadcast rule is similar to numpy.array(input) * numpy.ones(shape):
        Dimensions are right alignment;
        Two corresponding dimensions must have the same value, or one of them is equal to 1.
        Also, this operator is similar to numpy.broadcast_to(input, shape),
        but the major difference is numpy.broadcast_to() does not allow shape to be smaller than input.size().
        It is possible that the output.shape is not equal to shape, when some dimensions in shape is equal to 1,
        or the shape.ndim < input.shape.ndim.


        Args:
            input: Input tensor

            shape: A 1-D tensor indicates the shape you want to expand to, following the
                broadcast rule
        """

        schema = get_schema("Expand", 8, "")
        op = Op(self, "Expand", schema)
        return op(*self._prepare_inputs(schema, input, shape))

    T_Max = TypeVar("T_Max", DOUBLE, FLOAT, FLOAT16)

    def Max(self, *data_0: T_Max) -> T_Max:
        r"""[🌐 Max(8)](https://onnx.ai/onnx/operators/onnx__Max.html#max-8 "Online Documentation")


        Element-wise max of each of the input tensors (with Numpy-style broadcasting support).
        All inputs and outputs must have the same data type.
        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            data_0: (variadic) List of tensors for max.
        """

        schema = get_schema("Max", 8, "")
        op = Op(self, "Max", schema)
        return op(*self._prepare_inputs(schema, *data_0))

    T_MaxPool = TypeVar("T_MaxPool", DOUBLE, FLOAT, FLOAT16)

    I_MaxPool: TypeAlias = INT64

    def MaxPool(
        self,
        X: T_MaxPool,
        *,
        auto_pad: str = "NOTSET",
        kernel_shape: Sequence[int],
        pads: Optional[Sequence[int]] = None,
        storage_order: int = 0,
        strides: Optional[Sequence[int]] = None,
    ) -> Tuple[T_MaxPool, I_MaxPool]:
        r"""[🌐 MaxPool(8)](https://onnx.ai/onnx/operators/onnx__MaxPool.html#maxpool-8 "Online Documentation")


         MaxPool consumes an input tensor X and applies max pooling across
         the tensor according to kernel sizes, stride sizes, and pad lengths.
         max pooling consisting of computing the max on all values of a
         subset of the input tensor according to the kernel size and downsampling the
         data into the output tensor Y for further processing. The output spatial shape will be following:
         ```
         output_spatial_shape[i] = floor((input_spatial_shape[i] + pad_shape[i] - kernel_spatial_shape[i]) / strides_spatial_shape[i] + 1)

         * pad_shape[i] is sum of pads along axis i
         ```

         `auto_pad` is a DEPRECATED attribute. If you are using them currently, the output spatial shape will be following:
         ```
         VALID: output_spatial_shape[i] = ceil((input_spatial_shape[i] - kernel_spatial_shape[i] + 1) / strides_spatial_shape[i])
         SAME_UPPER or SAME_LOWER: output_spatial_shape[i] = ceil(input_spatial_shape[i] / strides_spatial_shape[i])
         ```
         And pad shape will be following if `SAME_UPPER` or `SAME_LOWER`:
         ```
         pad_shape[i] = (output_spatial_shape[i] - 1) * strides_spatial_shape[i] + kernel_spatial_shape[i] - input_spatial_shape[i]
         ```
         The output of each pooling window is maximum number of elements exclude pad.


        Args:
            X: Input data tensor from the previous operator; dimensions for image case
                are (N x C x H x W), where N is the batch size, C is the number of
                channels, and H and W are the height and the width of the data. For non
                image case, the dimensions are in the form of (N x C x D1 x D2 ... Dn),
                where N is the batch size. Optionally, if dimension denotation is in
                effect, the operation expects the input data tensor to arrive with the
                dimension denotation of [DATA_BATCH, DATA_CHANNEL, DATA_FEATURE,
                DATA_FEATURE ...].

            auto_pad: auto_pad must be either NOTSET, SAME_UPPER, SAME_LOWER or VALID.
                Where default value is NOTSET, which means explicit padding is used.
                SAME_UPPER or SAME_LOWER mean pad the input so that the output spatial
                size match the input.In case of odd number add the extra padding at the
                end for SAME_UPPER and at the beginning for SAME_LOWER. VALID mean no
                padding.

            kernel_shape: The size of the kernel along each axis.

            pads: Padding for the beginning and ending along each spatial axis, it can
                take any value greater than or equal to 0. The value represent the
                number of pixels added to the beginning and end part of the
                corresponding axis. `pads` format should be as follow [x1_begin,
                x2_begin...x1_end, x2_end,...], where xi_begin the number of pixels
                added at the beginning of axis `i` and xi_end, the number of pixels
                added at the end of axis `i`. This attribute cannot be used
                simultaneously with auto_pad attribute. If not present, the padding
                defaults to 0 along start and end of each spatial axis.

            storage_order: The storage order of the tensor. 0 is row major, and 1 is
                column major.

            strides: Stride along each spatial axis.
        """

        schema = get_schema("MaxPool", 8, "")
        op = Op(self, "MaxPool", schema)
        return op(
            *self._prepare_inputs(schema, X),
            auto_pad=auto_pad,
            kernel_shape=kernel_shape,
            pads=pads,
            storage_order=storage_order,
            strides=strides,
        )

    T_Mean = TypeVar("T_Mean", DOUBLE, FLOAT, FLOAT16)

    def Mean(self, *data_0: T_Mean) -> T_Mean:
        r"""[🌐 Mean(8)](https://onnx.ai/onnx/operators/onnx__Mean.html#mean-8 "Online Documentation")


        Element-wise mean of each of the input tensors (with Numpy-style broadcasting support).
        All inputs and outputs must have the same data type.
        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            data_0: (variadic) List of tensors for mean.
        """

        schema = get_schema("Mean", 8, "")
        op = Op(self, "Mean", schema)
        return op(*self._prepare_inputs(schema, *data_0))

    T_Min = TypeVar("T_Min", DOUBLE, FLOAT, FLOAT16)

    def Min(self, *data_0: T_Min) -> T_Min:
        r"""[🌐 Min(8)](https://onnx.ai/onnx/operators/onnx__Min.html#min-8 "Online Documentation")


        Element-wise min of each of the input tensors (with Numpy-style broadcasting support).
        All inputs and outputs must have the same data type.
        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            data_0: (variadic) List of tensors for min.
        """

        schema = get_schema("Min", 8, "")
        op = Op(self, "Min", schema)
        return op(*self._prepare_inputs(schema, *data_0))

    I_Scan: TypeAlias = INT64

    V_Scan = TypeVar(
        "V_Scan",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Scan(
        self,
        sequence_lens: Optional[I_Scan],
        *initial_state_and_scan_inputs: V_Scan,
        body: GraphProto,
        directions: Optional[Sequence[int]] = None,
        num_scan_inputs: int,
    ) -> V_Scan:
        r"""[🌐 Scan(8)](https://onnx.ai/onnx/operators/onnx__Scan.html#scan-8 "Online Documentation")


        Scan can be used to iterate over one or more scan_input tensors,
        constructing zero or more scan_output tensors. It combines ideas from general recurrences,
        functional programming constructs such as scan, fold, map, and zip, and is intended to enable
        generalizations of RNN-like constructs for sequence-to-sequence processing.
        Other tensors (referred to as state_variables here) can be used to carry a state
        when iterating from one element to another (similar to hidden-state in RNNs, also referred
        to as loop-carried dependences in the context of loops). All these tensors are required to
        have the same shape in each iteration of the loop (a restriction imposed to enable efficient
        memory allocation). Many common usages involve a single scan_input tensor (where functionality
        similar to scan, fold and map can be obtained). When more than one scan_input is used,
        a behavior similar to zip is obtained.

        The attribute body must be a graph, specifying the computation to be performed in
        every iteration. It takes as input the current values of the state_variables and
        the current iterated element of the scan_inputs. It must return the (updated) values
        of the state_variables and zero or more scan_output_element tensors. The values of the
        scan_output_element tensors are concatenated over all the iterations to produce the
        scan_output values of the scan construct (similar to the concatenated intermediate
        hidden-state values of RNN-like constructs).

        The scan operation returns the final values of the state_variables as well as the
        scan_outputs.

        The operation supports batching, and the batch-axis is required to be 0.
        When multiple scan_input tensors are used, they must all have the same batch-size,
        and they must all have the same maximum-sequence-length (the dimensionality of the
        sequence axis or scan axis). The sequence axis or scan axis is required to be 1.

        The operation has an optional sequence_lens input (of shape [BATCH_SIZE]) to
        allow variable length sequences of length <= the maximum-sequence-length. If this
        input is not specified, all sequences are assumed to be of length equal to
        maximum-sequence-length. For variable length input sequences, the scan_outputs
        will consist of a sequence of same length as the input, padded to the
        maximum-sequence-length.

        The optional attribute directions can be used to scan a sequence in the reverse direction.
        If this attribute is omitted, all sequences are scanned in the forward direction.
        A bidirectional scan be performed by specifying the same tensor input twice in the
        scan_inputs, once with a forward direction, and once with a backward direction.

        Note that because of the ONNX restriction that only the last parameter of an operator can
        be variadic, the initial-states and scan-inputs are listed together as one input parameter.
        Similarly, the final-states and scan-outputs are listed together as one output parameter.
        The attribute num_scan_inputs indicates the number M of scan-inputs.

        The behavior of

            Scan <
                num_scan_inputs = m,
                body = loop-body
            > (sequence_lengths, init_1, ..., init_n, scan_1, ..., scan_m)

        is equivalent to the following pseudo-code:

            // T.shape[0] denotes the batch-size of T
            // The batch-size of scan_1, ..., scan_m are all required to be equal
            batch_size = scan_1.shape[0];

            // scan_i.shape[1] denotes the (max) sequence-length of scan_i
            // scan_i.shape[1] is required to be equal to scan_j.shape[1] for all i,j.
            max_sequence_length = scan_1.shape[1];

            for (int batch = 0; batch < batch_size; ++batch) {
                // initialize state-variables
                st_1 = init_1; ... st_n = init_n;
                // initialize scan-output variables: [] denotes an empty tensor
                scan_out_1 = []; ...; scan_out_k = [];
                // identify number of iterations:
                N = (sequence_lengths specified) ? sequence_lengths[batch] : max_sequence_length;

                // execute loop
                for (int t = 0; t < N; ++t) {
                    // generate the scan-input elements: the notation T<axis=k>[t] indicates the sub-tensor
                    // of rank one less than T obtained by indexing T at position t along axis k.
                    si_1 = (scan_1<axis=0>[batch])<axis=1>[t];
                    ... ;
                    si_m = (scan_m<axis=0>[batch])<axis=1>[t];
                    // execute loop-body
                    st_1, ..., st_n, so_1, ..., so_k = loop-body(st_1, ..., st_n, si_1, ..., si_m)
                    // accumulate the scan-output elements
                    scan_out_1 = Concat<axis=0>(scan_out_1, so_1); ... ; scan_out_k = Concat<axis=0>(scan_out_k, so_k);
                }
                // accumulate the outputs for this batch:
                bst_1[batch] = st_1; ..., bst_n[batch] = st_n;
                // Note scan-outputs will have size max_sequence_length, but only first N values will be meaningful.
                // The remaining values have an undefined value.
                b_scan_out_1[batch] = scan_out_1; ...; b_scan_out_k[batch] = scan_out_k;
            }
            return bst_1, ..., bst_n, b_scan_out_1, ..., b_scan_out_k;



        *Sample usage: Encoding RNN using a Scan*

        The following example shows how a simple RNN over an input tensor %X, with weight tensor %Wi,
        recurrence weight tensor %Ri, bias tensors %Wbi and %Rbi, and initial hidden-state %H_0 can
        be encoded as a ScanLoop. Note that the loop-body is a nested graph, and it directly computes
        %Wi, %Ri, %Wbi, and %Rbi (typically constants or initializers in the body graph). If these
        values are computed in the outer graph, they need to be passed in as extra state_variables.

            graph rnn-encoding {
              %H_0 = ...
              %X = ...
              %Y_h, %Y = Scan[body = <graph rnn-cell-1>, num_scan_inputs=1]("", %H_0, %X)
              return %Y, %Y_h
            }

            graph rnn-cell-1 (
              %H_tminus1[FLOAT, tensor]
              %X_t[FLOAT, tensor]
            ) {
              %Wi = ...
              %Ri = ...
              %Wbi = ...
              %Rbi = ...
              %t1 = X_t * (Wi^T)
              %t2 = H_tminus1*(Ri^T)
              %t3 = Add(%t1, %t2)
              %t4 = Add(%t3, %Wbi)
              %t5 = Add(%t4, %Rbi)
              %Ht = Tanh(%t5)
              %Accumulate = Identity(%Ht)
              return %Ht, %Accumulate
            }



        Args:
            sequence_lens: (optional) Optional tensor specifying lengths of the
                sequences in a batch. If this input is not specified, all sequences are
                assumed to be of the maximum sequence length (the dimension of the
                sequence axis of the scan_input tensors).

            initial_state_and_scan_inputs: (variadic, heterogeneous) Initial values of
                the loop's N state variables followed by M scan_inputs

            body: The graph run each iteration. It has N+M inputs: (loop state
                variables..., scan_input_elts...). It has N+K outputs: (loop state
                variables..., scan_output_elts...). Each scan_output is created by
                concatenating the value of the specified scan_output_elt value at the
                end of each iteration of the loop. It is an error if the dimensions of
                these values change across loop iterations.

            directions: An optional list of M flags. The i-th element of the list
                specifies the direction to be scanned for the i-th scan_input tensor: 0
                indicates forward direction and 1 indicates reverse direction. If
                omitted, all scan_input tensors will be scanned in the forward
                direction.

            num_scan_inputs: An attribute specifying the number of scan_inputs M.
        """

        schema = get_schema("Scan", 8, "")
        op = Op(self, "Scan", schema)
        return op(
            *self._prepare_inputs(schema, sequence_lens, *initial_state_and_scan_inputs),
            body=body,
            directions=directions,
            num_scan_inputs=num_scan_inputs,
        )

    T_Sum = TypeVar("T_Sum", DOUBLE, FLOAT, FLOAT16)

    def Sum(self, *data_0: T_Sum) -> T_Sum:
        r"""[🌐 Sum(8)](https://onnx.ai/onnx/operators/onnx__Sum.html#sum-8 "Online Documentation")


        Element-wise sum of each of the input tensors (with Numpy-style broadcasting support).
        All inputs and outputs must have the same data type.
        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.


        Args:
            data_0: (variadic) List of tensors for sum.
        """

        schema = get_schema("Sum", 8, "")
        op = Op(self, "Sum", schema)
        return op(*self._prepare_inputs(schema, *data_0))



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from onnxscript.rewriter._rewrite_rule import RewriteRule
from onnxscript.rewriter.rules.common._broadcast_to_matmul import check_if_not_need_reshape


# Pattern to match against
def reshape_gemm_reshape_pattern(op, input_a, input_b, input_c, shape_a, shape_c):
    reshape_a = op.Reshape(input_a, shape_a)
    # TODO: Temporary workaround to support benchmodels.
    # Tracked by https://github.com/microsoft/onnx-rewriter/issues/197.
    gemm = op.Gemm(reshape_a, input_b, input_c, alpha=1.0, beta=1.0)
    return op.Reshape(gemm, shape_c)


def matmul_add(op, input_a, input_b, input_c, **_):
    matmul = op.MatMul(input_a, input_b)
    return op.Add(matmul, input_c)


gemm_to_matmul_add_rule = RewriteRule(
    reshape_gemm_reshape_pattern, matmul_add, check_if_not_need_reshape
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Convert the model to the specified ONNX opset version."""

from __future__ import annotations

import dataclasses
import functools
import logging
from typing import Callable, Sequence, Union

import onnx_ir.convenience as ir_convenience

import onnxscript.ir._tape as _tape
import onnxscript.utils.metadata_merger as metadata_merger
from onnxscript import ir

logger = logging.getLogger(__name__)


SUPPORTED_MAX_ONNX_OPSET = 25
SUPPORTED_MIN_ONNX_OPSET = 18


def _get_onnx_opset_version(model: ir.Model) -> int | None:
    """Get the ONNX opset version imported by the model."""
    model_version1 = model.opset_imports.get("")
    model_version2 = model.opset_imports.get("ai.onnx")
    if model_version1 is not None and model_version2 is not None:
        if model_version1 != model_version2:
            raise ValueError(
                f"Model imports multiple onnx opsets: {model_version1} and {model_version2}."
            )
    return model_version1 or model_version2


def _set_onnx_opset_version(model_or_function: ir.Model | ir.Function, version: int) -> None:
    """Set the ONNX opset version imported by the model or function."""
    if "ai.onnx" in model_or_function.opset_imports:
        del model_or_function.opset_imports["ai.onnx"]
    model_or_function.opset_imports[""] = version


class VersionConverterError(RuntimeError):
    """Raised when an node's version cannot be upgraded/downgraded successfully."""


@dataclasses.dataclass
class Replacement:
    """A replacement for a node in the graph."""

    new_outputs: Sequence[ir.Value]
    new_nodes: Sequence[ir.Node]


# A version-adapter function takes a node, a RewriterContext and returns
# a Replacement for the node or None (if no replacement is needed).

RewriterContext = _tape.Builder
ReturnValue = Union[Sequence[ir.Value], ir.Value, None]
AdapterFunction = Callable[[ir.Node, RewriterContext], ReturnValue]


def version_supported(model: ir.Model, target_version: int) -> bool:
    """Check if the target version is supported by the current version."""
    if "" in model.graph.opset_imports:
        current_version = model.graph.opset_imports[""]
    else:
        return True
    return (
        SUPPORTED_MIN_ONNX_OPSET
        <= current_version
        <= target_version
        <= SUPPORTED_MAX_ONNX_OPSET
    )


class AdapterRegistry:
    """A class that maintains a registry of adapters for ops."""

    def __init__(self):
        self.op_adapters: dict[tuple[str, str, int, bool], AdapterFunction] = {}

    def lookup_adapters(
        self,
        domain: str,
        opname: str,
        original_version: int,
        up_conversion: bool = True,
    ) -> AdapterFunction | None:
        adapter_func = self.op_adapters.get((domain, opname, original_version, up_conversion))
        if adapter_func is not None:
            return adapter_func
        return None

    def register(
        self, opname: str, domain: str = "", node_version=None, up_conversion=True
    ) -> Callable[[AdapterFunction], AdapterFunction]:
        """Register an adapter based on the domain, operator type, node version and whether to upgrade/downgrade node version"""

        def decorator(function: AdapterFunction) -> AdapterFunction:
            @functools.wraps(function)
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            self.op_adapters[(domain, opname, node_version, up_conversion)] = function
            return wrapped_function

        return decorator


registry: AdapterRegistry = AdapterRegistry()

register = registry.register


def _get_input(node: ir.Node, index: int) -> ir.Value | None:
    if index < len(node.inputs):
        return node.inputs[index]
    return None


def _get_int_attribute(node: ir.Node, name: str, default: int | None = None) -> int | None:
    if name in node.attributes:
        attr = node.attributes[name]
        if not isinstance(attr, ir.Attr):
            return None
        attr_val = attr.value
        if isinstance(attr_val, int):
            return attr_val
        # This is an invalid model: attribute has invalid/unexpected type.
        # For now, we just return None. We could raise an error too.
        return None
    return default


def _get_str_attribute(node: ir.Node, name: str, default: str | None = None) -> str | None:
    if name in node.attributes:
        attr = node.attributes[name]
        if not isinstance(attr, ir.Attr):
            return None
        attr_val = attr.value
        if isinstance(attr_val, str):
            return attr_val
        # This is an invalid model: attribute has invalid/unexpected type.
        # For now, we just return None. We could raise an error too.
        return None
    return default


## Op-specific adapters

# Opset 19 -> 20


@register("DFT", node_version=19, up_conversion=True)
def dft_19_20(node: ir.Node, op):
    input = node.inputs[0]
    dft_length = node.inputs[1] if len(node.inputs) > 1 else None
    inverse = _get_int_attribute(node, "inverse", 0)
    onesided = _get_int_attribute(node, "onesided", 0)
    axis = _get_int_attribute(node, "axis", None)
    if axis is not None:
        axis_value = op.Constant(value_int=axis)
        return op.DFT(input, dft_length, axis_value, inverse=inverse, onesided=onesided)
    return None


@register("GridSample", node_version=19, up_conversion=True)
def gridsample_19_20(node: ir.Node, op):
    x = node.inputs[0]
    grid = node.inputs[1]
    align_corners = _get_int_attribute(node, "align_corners", 0)
    mode = _get_str_attribute(node, "mode", "linear")
    padding_mode = _get_str_attribute(node, "padding_mode", "zeros")
    if mode == "bilinear":
        return op.GridSample(
            x, grid, align_corners=align_corners, mode="linear", padding_mode=padding_mode
        )
    elif mode == "bicubic":
        return op.GridSample(
            x, grid, align_corners=align_corners, mode="cubic", padding_mode=padding_mode
        )
    return None


# Opset 20 -> 21


@register("GroupNormalization", node_version=20, up_conversion=True)
def groupnormalization_20_21(node: ir.Node, op):
    x = _get_input(node, 0)
    scale = _get_input(node, 1)
    bias = _get_input(node, 2)
    if x is None or scale is None or bias is None:
        raise VersionConverterError(f"Missing input for {node}")

    x_shape = x.shape
    if x_shape is None:
        raise VersionConverterError(f"Missing required shape for {x}")
    num_channels = x_shape[1]
    if not isinstance(num_channels, int):
        return None

    scale_shape = scale.shape
    bias_shape = bias.shape
    if scale_shape is None or bias_shape is None:
        return None
    if not isinstance(scale_shape[0], int) or not isinstance(bias_shape[0], int):
        return None

    num_groups = _get_int_attribute(node, "num_groups", None)
    if num_groups is None:
        raise VersionConverterError("Missing required attribute: num_groups")
    if (
        num_groups != num_channels
        and num_groups == scale_shape[0]
        and num_groups == bias_shape[0]
    ):
        reshape_1_sizes = op.Constant(value_ints=[-1, 1])
        reshape_2_sizes = op.Constant(value_ints=[-1])
        c_div = int(num_channels / num_groups)
        expand_sizes = op.Constant(value_ints=[1, c_div])

        # Modify scale input
        scale_reshape_1 = op.Reshape(scale, reshape_1_sizes)
        scale_expand = op.Expand(scale_reshape_1, expand_sizes)
        scale_reshape_2 = op.Reshape(scale_expand, reshape_2_sizes)

        # Modify bias input
        bias_reshape_1 = op.Reshape(bias, reshape_1_sizes)
        bias_expand = op.Expand(bias_reshape_1, expand_sizes)
        bias_reshape_2 = op.Reshape(bias_expand, reshape_2_sizes)

        return op.GroupNormalization(x, scale_reshape_2, bias_reshape_2, num_groups=num_groups)
    return None


class _VersionConverter:
    def __init__(self, target_version: int):
        self._target_version = target_version
        # Default metadata merger: no merging should be needed; keep the first value.
        self._default_metadata_merger: metadata_merger.MetadataMerger = (
            metadata_merger.MetadataMerger(
                dict(),
            )
        )

    def process_node(
        self, node: ir.Node, from_version: int, up_conversion: bool = True
    ) -> Replacement | None:
        assert node.domain == ""
        adapter = registry.lookup_adapters(
            node.domain, node.op_type, from_version, up_conversion
        )
        if adapter is None:
            return None
        context = RewriterContext()
        output = adapter(node, context)
        if output is not None:
            if isinstance(output, ir.Value):
                output = [output]
            return Replacement(output, context.nodes)
        return None

    def replace_node(self, node: ir.Node, replacement, root: ir.Graph | ir.Function) -> None:
        logger.debug("Replacing node: %s::%s %s", node.domain, node.op_type, node.name)

        ir_convenience.replace_nodes_and_values(
            root, node, [node], replacement.new_nodes, node.outputs, replacement.new_outputs
        )

    def visit_attribute(self, attr: ir.Attr) -> None:
        if attr.is_ref():
            return
        if attr.type == ir.AttributeType.GRAPH:
            self.visit_graph_or_function(attr.as_graph())
        elif attr.type == ir.AttributeType.GRAPHS:
            for graph in attr.as_graphs():
                self.visit_graph_or_function(graph)

    def visit_node(
        self,
        node: ir.Node,
        root: ir.Graph | ir.Function,
        from_version: int,
        up_conversion: bool = True,
    ) -> None:
        if up_conversion:
            to_version = from_version + 1
        else:
            to_version = from_version - 1
        replacement = self.process_node(node, from_version, up_conversion)
        if replacement is None:
            # No change. Process attributes.
            for attr in node.attributes.values():
                self.visit_attribute(attr)
            node.version = to_version
        else:
            for new_node in replacement.new_nodes:
                # TODO: control-flow
                new_node.version = to_version
            self._default_metadata_merger.copy_merged_metadata([node], replacement.new_nodes)
            self.replace_node(node, replacement, root)

    def visit_graph_or_function(self, graph_or_function: ir.Graph | ir.Function) -> None:
        for node in graph_or_function:
            if node.domain != "":
                continue
            node_version = node.version or self._default_onnx_opset
            if node_version is None:
                raise VersionConverterError(f"Node {node} has no version.")
            # RefAttr is not supported by adapters for now.
            if any(attr.is_ref() for attr in node.attributes.values()):
                raise VersionConverterError(
                    f"Node '{node!r}' has ref attribute, which is not supported by version converter."
                )
            # Iterate each node from current node version -> target version
            # and updating node based on the correct adapter
            # Up-conversion [ver->ver+1] or down-conversion [ver->ver-1]
            # TODO(shubhambhokare1): Remove once down-conversion adapters are supoorted
            if self._target_version < node_version:
                raise VersionConverterError(
                    f"Target opset: {self._target_version} less than node version: {node.version}, "
                    "downstream version conversion not currently handled."
                )
            for from_version in range(node_version, self._target_version):
                try:
                    self.visit_node(node, graph_or_function, from_version, up_conversion=True)
                except VersionConverterError as e:
                    logger.warning(
                        "Skipping version conversion for node %s due to exception: %s",
                        node.op_type,
                        e,
                    )

    def visit_model(self, model: ir.Model) -> None:
        self._default_onnx_opset = _get_onnx_opset_version(model)
        self.visit_graph_or_function(model.graph)
        for function in model.functions.values():
            self.visit_graph_or_function(function)
            _set_onnx_opset_version(function, self._target_version)
        _set_onnx_opset_version(model, self._target_version)


def convert_version(model: ir.Model, target_version: int) -> None:
    """Convert the model to the specified ONNX opset version."""
    if (target_version > SUPPORTED_MAX_ONNX_OPSET) or (
        target_version < SUPPORTED_MIN_ONNX_OPSET
    ):
        raise ValueError(
            f"Target opset version {target_version} is not supported. "
            f"Supported range: {SUPPORTED_MIN_ONNX_OPSET} to {SUPPORTED_MAX_ONNX_OPSET}."
        )
    version_converter = _VersionConverter(target_version=target_version)
    version_converter.visit_model(model)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Does the following transformation:
- Relu(Relu(X)) -> Relu
- Relu(Clip(X)) -> Clip
- Clip(Relu(X)) -> Clip
- Clip(Clip(X)) -> Clip
"""

from __future__ import annotations

import abc

import numpy as np
import onnx_ir as ir

from onnxscript.rewriter._basics import MatchResult
from onnxscript.rewriter._rewrite_rule import RewriteRuleClassBase, RewriteRuleSet


class FuseSuccessiveRelu(RewriteRuleClassBase):
    """Replaces ``Relu(Relu(X))`` with ``Relu(X)``."""

    def rewrite(self, op, x):
        return op.Relu(x)

    def pattern(self, op, x):
        return op.Relu(op.Relu(x))


class _FuseReluClipBase(RewriteRuleClassBase, abc.ABC):
    def rewrite(self, op, x, **kwargs):
        first_clip_node = kwargs.get("out_first_clip").producer()
        second_clip_node = None

        if out_second_clip := kwargs.get("out_second_clip"):
            second_clip_node = out_second_clip.producer()

        min_clip, max_clip = self.compute_clip_min_max(first_clip_node, second_clip_node)
        clip_min_max = []

        if min_clip is not None:
            clip_min_max.append(
                op.initializer(min_clip, name=f"{first_clip_node.inputs[0].name}_min")
            )

        if max_clip is not None:
            # ONNX Clip expects min and max inputs in order.
            # If min is not provided, we insert None to maintain correct argument positions.
            if min_clip is None:
                clip_min_max.append(None)

            clip_min_max.append(
                op.initializer(max_clip, name=f"{first_clip_node.inputs[0].name}_max")
            )

        return op.Clip(x, *clip_min_max)

    @abc.abstractmethod
    def compute_clip_min_max(
        self, first_clip_node: ir.Node, second_clip_node: ir.Node | None = None
    ):
        pass

    def extract_min_max(self, node: ir.Node):
        # Infer dtype from node first input
        dtype = node.inputs[0].dtype.numpy()
        min_clip, max_clip = None, None

        if len(node.inputs) > 1:
            min_input = node.inputs[1]
            # If only a max is provided, min is implicitly None, so we check that
            if min_input is not None:
                min_clip = min_input.const_value.numpy()

        if len(node.inputs) > 2:
            max_clip = node.inputs[2].const_value.numpy()

        return min_clip, max_clip, dtype

    def check(self, context, **kwargs):
        """Condition to check if we need to replace the pattern.

        The pattern is applied only when the min and max inputs of the Clip nodes are
        not graph inputs and are constant values (i.e., provided by Constant nodes or initializers).

        Returns:
            MatchResult:
                Success if we need to replace the pattern, Failure otherwise.
        """
        del context  # Unused
        check_result = MatchResult()

        # Check if Clip min/max are not graph inputs and are constant values
        clip_min_max = []

        first_clip_node = kwargs.get("out_first_clip").producer()
        clip_min_max.extend([inp for inp in first_clip_node.inputs[1:] if inp is not None])

        if out_second_clip := kwargs.get("out_second_clip"):
            second_clip_node = out_second_clip.producer()
            clip_min_max.extend(
                [inp for inp in second_clip_node.inputs[1:] if inp is not None]
            )

        for m in clip_min_max:
            if m.is_graph_input():
                return check_result.fail(f"{m.name} is a graph input.")

            if ir.convenience.get_const_tensor(m) is None:
                return check_result.fail(f"{m.name} is not a constant.")

        return check_result


class FuseSuccessiveClip(_FuseReluClipBase):
    """Replaces ``Clip(Clip(X))`` with ``Clip(X)``."""

    def pattern(self, op, x):
        return op.Clip(
            op.Clip(x, _allow_other_inputs=True, _outputs=["out_first_clip"]),
            _allow_other_inputs=True,
            _outputs=["out_second_clip"],
        )

    def compute_clip_min_max(self, first_clip_node: ir.Node, second_clip_node: ir.Node):
        min_clip1, max_clip1, dtype = self.extract_min_max(first_clip_node)
        min_clip2, max_clip2, _ = self.extract_min_max(second_clip_node)

        def combine(val1, val2, op):
            if val1 is not None and val2 is not None:
                return ir.tensor(np.array(op(val1, val2), dtype=dtype))
            elif val1 is not None:
                return ir.tensor(val1)
            elif val2 is not None:
                return ir.tensor(val2)
            return None

        min_clip = combine(min_clip1, min_clip2, np.maximum)
        max_clip = combine(max_clip1, max_clip2, np.minimum)

        return min_clip, max_clip


class FuseSuccessiveClipRelu(_FuseReluClipBase):
    """Replaces ``Clip(Relu(X))`` with ``Clip(X)``."""

    def pattern(self, op, x):
        return op.Clip(op.Relu(x), _allow_other_inputs=True, _outputs=["out_first_clip"])

    def compute_clip_min_max(self, first_clip_node: ir.Node, _):
        min_clip, max_clip, dtype = self.extract_min_max(first_clip_node)

        if min_clip is None:
            # The minimum clipping value is implicitly 0 (Relu clamps at 0)
            min_clip = 0

        min_clip = ir.tensor(np.array(np.maximum(0.0, min_clip), dtype=dtype))

        if max_clip is not None:
            max_clip = ir.tensor(max_clip)
        return min_clip, max_clip


class FuseSuccessiveReluClip(FuseSuccessiveClipRelu):
    """Replaces ``Relu(Clip(X))`` with ``Clip(X)``."""

    def pattern(self, op, x):
        return op.Relu(op.Clip(x, _allow_other_inputs=True, _outputs=["out_first_clip"]))


successive_relu_rule = FuseSuccessiveRelu().rule()
successive_clip_rule = FuseSuccessiveClip().rule()
successive_clip_relu_rule = FuseSuccessiveClipRelu().rule()
successive_relu_clip_rule = FuseSuccessiveReluClip().rule()


rules = RewriteRuleSet(
    [
        successive_clip_relu_rule,
        successive_relu_clip_rule,
        successive_relu_rule,
        successive_clip_rule,
    ]
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Fuses Pad nodes into preceding nodes. Supported fusion patterns:
- Conv ∘ Pad          -> Conv
- ConvInteger ∘ Pad   -> ConvInteger

To make some rules possible, we implicitly transform `auto_pad` attribute into its explicit list.
"""

from __future__ import annotations

from typing import List, Sequence

import numpy as np
import onnx_ir as ir

from onnxscript.rewriter import pattern as orp


def fill_pads_with_axes(pads: Sequence[int], axes: Sequence[int], rank: int) -> List[int]:
    """Converts the parameters of the ONNX Pad operator into an explicit list of values.

    A filled list of pads will be returned following the format:
    [x1_begin, x2_begin, ..., x{rank}_begin, x1_end, x2_end, ..., x{rank}_end]

    Args:
        pads: list of integers indicating the number of padding elements to add at
            the beginning and end of each axis.
        axes: list of axes that pads apply to.
        rank: value to compute the size of the filled list (2 * rank).

    Returns:
        The filled list of pads.
    """
    new_pads = [0] * 2 * rank
    N = len(axes)
    for start_idx, axis in enumerate(axes):
        new_pads[axis] = pads[start_idx]
        new_pads[axis + rank] = pads[start_idx + N]
    return new_pads


def read_conv_attributes(ir_conv: ir.Node) -> dict[str, Sequence[int] | str]:
    # Read attributes
    attributes = {}
    ir_attributes = ir_conv.attributes
    attributes["kernel_shape"] = ir_attributes.get_ints(
        "kernel_shape", ir_conv.inputs[1].shape[2:]
    )
    attributes["strides"] = ir_attributes.get_ints(
        "strides", [1] * len(ir_conv.inputs[0].shape[2:])
    )
    attributes["auto_pad"] = ir_attributes.get_string("auto_pad", "NOTSET")
    if "pads" in ir_attributes:
        attributes["pads"] = ir_attributes.get_ints("pads")
    return attributes


class _FuseConvPadBase(orp.RewriteRuleClassBase):
    """Interface for PadConv nodes fusion."""

    def __init__(self, as_function: bool = False):
        # Remove nodes is set to False to remove unused nodes after the rewrite, since
        # Pad or Conv inputs can come from constant nodes.
        # With remove_nodes=False these nodes are removed if these nodes are no longer needed.
        super().__init__(remove_nodes=False, as_function=as_function)

    def rewrite(
        self, op: ir.tape.Tape, x: ir.Value, pad: ir.Value, conv: ir.Value
    ) -> ir.Value:
        conv_node = conv.producer()

        # Retrieve the padding and axes
        x_rank = len(x.shape)

        # Get computed pads in check()
        pad_pads = self._pads_list

        # Get only spatial pads
        new_pads = pad_pads[2:x_rank] + pad_pads[x_rank + 2 :]

        # Replace conv pads = new + old
        conv_attr = conv_node.attributes.copy()
        if "pads" in conv_attr:
            new_pads = [x + y for x, y in zip(conv_attr["pads"].as_ints(), new_pads)]
        conv_attr.add(ir.AttrInt64s("pads", new_pads))

        return op.op(
            conv_node.op_type,
            inputs=(x, *conv_node.inputs[1:]),
            attributes=conv_attr,
            domain=conv_node.domain,
            name=conv_node.name,
        )

    def check(self, context, x: ir.Value, pad: ir.Value, conv: ir.Value) -> orp.MatchResult:
        """Condition to check if we need to replace the pattern.

        If Pad inputs can be added in 'pads' attribute of the Conv operator.

        To validate this, we need to check the following:
        1. `Pad<mode>` attribute has 'constant' as value
        2. `Pad` operator inputs are constants ('pads', 'constant_value', 'axes')
        3. 'constant_value' is equal to 0.0.
        4. `Pad` operator is only used for the spatial dimensions (batch dimension and channels
           remain unchanged).

        If the above are true, then we don't need the reshapes.

        Returns:
            True if we need to replace the pattern, False otherwise.
        """
        del context  # Unused
        check_result = orp.MatchResult()
        pad_node = pad.producer()
        if x.shape is None:
            return check_result.fail(
                f"Input shapes are not defined on {pad_node.name} ({pad_node.op_type})."
            )
        x_rank = len(x.shape)

        # Pad constraints: attributes
        if (mode := pad_node.attributes.get("mode", None)) and mode.as_string() != "constant":
            return check_result.fail(
                f"{pad_node.name} ({pad_node.op_type}) mode must be 'constant'."
            )

        # Pad constraints: inputs
        if (pads := pad_node.inputs[1]).const_value is None:
            return check_result.fail(f"{pads.name} is not a constant/initializer.")
        if len(pad_node.inputs) > 2 and (constant_value := pad_node.inputs[2]) is not None:
            if constant_value.const_value is None:
                return check_result.fail(
                    f"{constant_value.name} is not a constant/initializer."
                )
            elif constant_value.const_value.numpy().item() != 0:
                return check_result.fail(f"{constant_value.name} must be equal to 0.")
        if len(pad_node.inputs) > 3 and (axes := pad_node.inputs[3]) is not None:
            if axes.const_value is None:
                return check_result.fail(f"{axes.name} is not a constant/initializer.")
            axes_list = [x if x >= 0 else x_rank + x for x in axes.const_value.numpy()]
        else:
            axes_list = list(range(x_rank))

        # Pad constraints: values
        self._pads_list = fill_pads_with_axes(pads.const_value.numpy(), axes_list, x_rank)
        if np.any(self._pads_list[:2] + self._pads_list[x_rank : x_rank + 2]):
            self._pads_list = None
            return check_result.fail(f"{pads.name} must be zero in non-spatial dimensions.")

        return check_result


class FuseConvPad(_FuseConvPadBase):
    """Replaces ``Conv(Pad(x))`` with ``Conv(x)``."""

    def pattern(self, op: ir.tape.Tape, x: ir.Value) -> ir.Value:
        return op.Conv(
            op.Pad(x, _allow_other_inputs=True, _outputs=["pad"]),
            _allow_other_inputs=True,
            _outputs=["conv"],
        )

    def check(self, context, x: ir.Value, pad: ir.Value, conv: ir.Value) -> orp.MatchResult:
        check_result = super().check(context, x, pad, conv)
        if not check_result:
            return check_result

        # Conv constraints: attributes
        conv_node = conv.producer()
        if conv_node.attributes.get_string("auto_pad", "NOTSET") != "NOTSET":
            return check_result.fail(
                f"{conv_node.name} ({conv_node.op_type}) auto_pad must be 'NOTSET'."
            )
        return check_result


class FuseConvIntegerPad(FuseConvPad):
    """Replaces ``ConvInteger(Pad(x))`` with ``ConvInteger(x)``."""

    def pattern(self, op: ir.tape.Tape, x: ir.Value) -> ir.Value:
        return op.ConvInteger(
            op.Pad(x, _allow_other_inputs=True, _outputs=["pad"]),
            _allow_other_inputs=True,
            _outputs=["conv"],
        )


class _NormalizePadFormatBase(orp.RewriteRuleClassBase):
    """Interface to normalize pad attributes in conv nodes."""

    @staticmethod
    def compute_pads(
        input_shape: Sequence[int],
        output_shape: Sequence[int],
        attributes: dict[str, Sequence[int] | str],
    ) -> Sequence[int]:
        raise NotImplementedError("Child have to implement this function")

    def rewrite(self, op: ir.tape.Tape, conv: ir.Value, **__) -> ir.Value:
        conv_node = conv.producer()

        # Read spatial dimensions and attributes
        input_shape = conv_node.inputs[0].shape[2:]
        output_shape = conv_node.outputs[0].shape[2:]
        attributes = read_conv_attributes(conv_node)

        # Convert auto_pad mode into an explicit list
        pads = self.compute_pads(input_shape, output_shape, attributes)

        # Replace auto_pad, forcing to the explicit list
        conv_attr = conv_node.attributes.copy()
        conv_attr.add(ir.AttrString("auto_pad", "NOTSET"))
        if any(x != 0 for x in pads):
            conv_attr.add(ir.AttrInt64s("pads", pads))

        return op.op(
            conv_node.op_type,
            inputs=conv_node.inputs,
            attributes=conv_attr,
            domain=conv_node.domain,
            name=conv_node.name,
        )

    def check(self, context, conv: ir.Value, **__) -> orp.MatchResult:
        """Condition to check if we need to replace the pattern.

        If it is possible to deduce 'pads'.

        To validate this, we need to check the following:
        1. `Conv<auto_pad != "NOTSET">` (nothing to do in this case, since 'pads' are
           already explicit)
        2. it is possible to deduce the input rank when `Conv<auto_pad == "VALID">`
        3. When `Conv<auto_pad != "VALID">`:
            * spatial input/output shapes are static
            * it is possible to infer `kernel_shape` either from the `Conv` operator attribute
              or from the kernel input

        If the above are true, then we don't need the reshapes.

        Returns:
            True if we need to replace the pattern, False otherwise.
        """
        del context
        check_result = orp.MatchResult()

        # Conv constraints: attributes
        conv_node = conv.producer()
        auto_pad = conv_node.attributes.get_string("auto_pad", None)
        if auto_pad in {None, "NOTSET"}:
            return check_result.fail(
                f"{conv_node.name} ({conv_node.op_type}) auto_pad must be different to 'NOTSET'."
            )

        # Conv constraints: inputs/outputs
        input_shape = conv_node.inputs[0].shape
        output_shape = conv_node.outputs[0].shape
        if input_shape is None or len(input_shape) <= 2:
            return check_result.fail(
                f"Input shapes are not defined on {conv_node.name} ({conv_node.op_type})."
            )
        if output_shape is None or len(output_shape) <= 2:
            return check_result.fail(
                f"Output shapes are not defined on {conv_node.name} ({conv_node.op_type})."
            )

        # Conv constraints: values
        if auto_pad != "VALID":
            error_msg = (
                "Expected static spatial {} shapes on "
                + conv_node.name
                + f" ({conv_node.op_type})."
            )
            if not all(isinstance(x, int) for x in input_shape[2:]):
                return check_result.fail(error_msg.format("input"))
            if not all(isinstance(x, int) for x in output_shape[2:]):
                return check_result.fail(error_msg.format("output"))
            attributes = read_conv_attributes(conv_node)
            if len(attributes["kernel_shape"]) != len(attributes["strides"]):
                return check_result.fail(
                    "strides must have the same length than kernel_shape on "
                    f"{conv_node.name} ({conv_node.op_type})."
                )
        return check_result


class NormalizePadFormatConv(_NormalizePadFormatBase):
    """Convert auto_pad attribute into 'NOTSET' in Conv nodes ."""

    @staticmethod
    def compute_pads(
        input_shape: Sequence[int],
        output_shape: Sequence[int],
        attributes: dict[str, Sequence[int] | str],
    ) -> Sequence[int]:
        # Compute pads, following auto_pad/pads attributes
        if attributes["auto_pad"] in {"NOTSET", "VALID"}:
            assert len(input_shape) > 0
            return attributes.get("pads", [0] * len(input_shape) * 2)

        bottom_pads, top_pads = [], []
        kernel_shape, strides = attributes["kernel_shape"], attributes["strides"]
        assert len(kernel_shape) == len(strides) == len(input_shape) == len(output_shape)
        for x, y, k, s in zip(input_shape, output_shape, kernel_shape, strides):
            # Compute the output shape and the total padding to apply
            total_pads = max(0, (y - 1) * s + k - x)

            # Depending of mode, apply the padding to the upper or lower part
            pad1 = total_pads // 2
            pad2 = total_pads - pad1
            if attributes["auto_pad"] == "SAME_UPPER":
                bottom_pads.append(pad1)
                top_pads.append(pad2)
            else:
                top_pads.append(pad1)
                bottom_pads.append(pad2)
        return bottom_pads + top_pads

    def pattern(self, op: ir.tape.Tape, x: ir.Value) -> ir.Value:
        return op.Conv(x, _allow_other_inputs=True, _outputs=["conv"])


class NormalizePadFormatConvInteger(NormalizePadFormatConv):
    """Convert auto_pad attribute into 'NOTSET' in ConvInteger nodes ."""

    def pattern(self, op: ir.tape.Tape, x: ir.Value) -> ir.Value:
        return op.ConvInteger(x, _allow_other_inputs=True, _outputs=["conv"])


normalize_pad_format_conv_rule = NormalizePadFormatConv.rule()
normalize_pad_format_conv_integer_rule = NormalizePadFormatConvInteger.rule()
fuse_pad_into_conv_rule = FuseConvPad.rule()
fuse_pad_into_conv_integer_rule = FuseConvIntegerPad.rule()


rules = orp.RewriteRuleSet(
    [
        normalize_pad_format_conv_rule,
        normalize_pad_format_conv_integer_rule,
        fuse_pad_into_conv_rule,
        fuse_pad_into_conv_integer_rule,
    ]
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from typing import TypeVar

__all__ = [
    "basic_constant_propagation",
    "fold_constants_ir",
    "fold_constants",
    "FOLDED_FROM_KEY",
    "inline",
    "optimize_ir",
    "optimize",
    "remove_unused_nodes",
]

import onnx
import onnx_ir.passes.common as common_passes

import onnxscript.optimizer._constant_folding as constant_folding
from onnxscript import ir
from onnxscript.optimizer._constant_folding import FOLDED_FROM_KEY, basic_constant_propagation
from onnxscript.optimizer._constant_folding import fold_constants as fold_constants_ir
from onnxscript.optimizer._optimizer import optimize_ir

_ModelProtoOrIr = TypeVar("_ModelProtoOrIr", onnx.ModelProto, ir.Model)


def optimize(
    model: _ModelProtoOrIr,
    num_iterations: int = 2,
    *,
    onnx_shape_inference: bool = True,
    stop_if_no_change: bool = True,
    input_size_limit: int = constant_folding.DEFAULT_CONSTANT_FOLD_INPUT_SIZE_LIMIT,
    output_size_limit: int = constant_folding.DEFAULT_CONSTANT_FOLD_OUTPUT_SIZE_LIMIT,
    inline: bool = True,
) -> _ModelProtoOrIr:
    """Optimizes a model.

    Args:
        model: The model to be optimized.
        num_iterations: Number of times the optimization loop is repeated.
        onnx_shape_inference: Applies node-level shape-inference as part of optimization
        input_size_limit: Will not apply constant folding to ops with any input of size
            greater than this. Does not apply to special ops like Shape() and Size().
        output_size_limit: Will not rewrite any foldable-op into a Constant op if the size
            of the output tensor is greater than this.
        stop_if_no_change: Stop the optimization loop if no change is detected in an iteration.
        inline: If True, inlines all functions in the model.

    Returns:
        The optimized model. If the input was a ModelProto, the output will also be a
        ModelProto. If the input was an ir.Model, the output will also be an ir.Model.
    """
    if isinstance(model, ir.Model):
        # In this case, optimize is done inplace.
        # TODO(justinchuby): Maybe make functional
        optimize_ir(
            model,
            num_iterations=num_iterations,
            onnx_shape_inference=onnx_shape_inference,
            stop_if_no_change=stop_if_no_change,
            input_size_limit=input_size_limit,
            output_size_limit=output_size_limit,
            inline=inline,
        )
        return model
    else:
        assert isinstance(model, onnx.ModelProto)
        model_ir = ir.serde.deserialize_model(model)
        optimize_ir(
            model_ir,
            num_iterations=num_iterations,
            onnx_shape_inference=onnx_shape_inference,
            stop_if_no_change=stop_if_no_change,
            input_size_limit=input_size_limit,
            output_size_limit=output_size_limit,
            inline=inline,
        )
        # Move the model back to the proto
        new_proto = ir.serde.serialize_model(model_ir)
        return new_proto


def inline(model: ir.Model) -> None:
    """Inline all function calls (recursively) in the model."""
    if model.functions:
        common_passes.InlinePass()(model)


def fold_constants(
    model: ir.Model | onnx.ModelProto, *args, **kwargs
) -> constant_folding.FoldConstantsResult:
    """Fold constants in a model in place."""
    if isinstance(model, ir.Model):
        return constant_folding.fold_constants(model, *args, **kwargs)
    else:
        assert isinstance(model, onnx.ModelProto)
        model_proto = model
        model = ir.serde.deserialize_model(model_proto)
        result = constant_folding.fold_constants(model, *args, **kwargs)
        # Move the model back to the proto
        new_proto = ir.serde.serialize_model(model)
        model_proto.Clear()
        model_proto.CopyFrom(new_proto)
        return result


def remove_unused_nodes(model: ir.Model | onnx.ModelProto) -> None:
    """Removes unused nodes from a model inplace."""
    if isinstance(model, ir.Model):
        common_passes.RemoveUnusedNodesPass()(model)
    else:
        model_ir = ir.serde.deserialize_model(model)
        model_ir = common_passes.RemoveUnusedNodesPass()(model_ir).model
        new_proto = ir.serde.serialize_model(model_ir)
        model.Clear()
        model.CopyFrom(new_proto)


def remove_unused_functions(model: ir.Model | onnx.ModelProto) -> None:
    """Removes unused functions from a model inplace."""
    if isinstance(model, ir.Model):
        common_passes.RemoveUnusedFunctionsPass()(model)
    else:
        model_ir = ir.serde.deserialize_model(model)
        model_ir = common_passes.RemoveUnusedFunctionsPass()(model_ir).model
        new_proto = ir.serde.serialize_model(model_ir)
        model.Clear()
        model.CopyFrom(new_proto)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import onnx_ir as ir

from onnxscript.rewriter.rules.fusion import _gqa, _rms_normalization, _rotary_embedding


def _get_onnx_opset_version(model: ir.Model) -> int | None:
    """Get the ONNX opset version imported by the model."""
    model_version1 = model.opset_imports.get("")
    model_version2 = model.opset_imports.get("ai.onnx")
    if model_version1 is not None and model_version2 is not None:
        if model_version1 != model_version2:
            raise ValueError(
                f"Model imports multiple onnx opsets: {model_version1} and {model_version2}."
            )
    return model_version1 or model_version2


def _opset_23_fuse(model: ir.Model, *, debug: bool = False) -> dict[str, int]:
    """Apply fusions targeting ONNX opset 23."""
    counts: dict[str, int] = {}
    counts["RMSNormalization"] = _rms_normalization.fuse_rms_normalization(model, debug=debug)
    counts["RotaryEmbedding"] = _rotary_embedding.fuse_rotary_embedding(model, debug=debug)
    counts["GQA"] = _gqa.fuse_gqa(model, debug=debug)
    return counts


def fuse(model: ir.Model, *, debug: bool = False) -> dict[str, int]:
    """Apply fusions targeting ONNX ops."""
    model_opset_version = _get_onnx_opset_version(model)
    if model_opset_version == 23:
        return _opset_23_fuse(model, debug=debug)
    return {}



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Rewrite rules for ONNX models."""

from __future__ import annotations

import abc
import dataclasses
import itertools
from typing import (
    Callable,
    Sequence,
    TypeVar,
)

import onnxscript.optimizer
import onnxscript.rewriter._basics as _basics
import onnxscript.rewriter._ir_utils as _ir_utils
import onnxscript.rewriter._matcher as _matcher
import onnxscript.rewriter._pattern_ir as _pattern_ir
import onnxscript.utils.metadata_merger as metadata_merger
from onnxscript import ir
from onnxscript.ir import _tape, convenience

T = TypeVar("T")

RewriterContext = _tape.Builder

# TODO(rama): Standardize metadata property keys. May be worth standardizing at ONNX level for
# source/producer metadata.

RULE_NAME_TAG = "pkg.onnxscript.rewriter.rule_name"


@dataclasses.dataclass
class ReplacementSubgraph:
    """A subgraph that will replace the matched pattern."""

    match: _basics.MatchResult
    new_outputs: Sequence[ir.Value]
    new_nodes: Sequence[ir.Node]
    new_initializers: Sequence[ir.Value]
    used_opsets: _tape.UsedOpsets


def always_true(*args, **kwargs) -> bool:
    """A condition function that always returns True.

    This is used when no condition function is provided for a rewrite rule.
    """
    return True


class Pattern:
    """A pattern that can be matched against nodes in an ONNX graph.

    This class encapsulates pattern matching functionality, providing the ability to
    match patterns against nodes without requiring replacement functionality.
    """

    def __init__(
        self,
        target_pattern: _pattern_ir.GraphPattern | Callable,
        condition_function: Callable | None = None,
        matcher: _matcher.PatternMatcher
        | Callable[[_pattern_ir.GraphPattern], _matcher.PatternMatcher]
        | None = None,
        verbose: int = 0,
        name: str | None = None,
    ) -> None:
        """Create a pattern matcher.

        Args:
            target_pattern: The _pattern_ir.GraphPattern that will be matched against the IR.
                If a callable is provided, it will be converted to a _pattern_ir.GraphPattern.
            condition_function: The condition function that will be used to check if
                the pattern match found should be rewritten.
            matcher: The pattern matcher that will be used to match the pattern.
                If not provided, a default matcher will be used.
            verbose: The verbosity level of the rule.
            name: An optional name for the pattern that will show up in verbose logging.
        """
        if not isinstance(target_pattern, _pattern_ir.GraphPattern):
            target_pattern = _pattern_ir._to_graph_pattern(target_pattern)
        self._target_pattern = target_pattern

        self._condition_function = condition_function or always_true
        if isinstance(matcher, _matcher.PatternMatcher):
            self._matcher = matcher
        elif matcher is None:
            self._matcher = _matcher.SimplePatternMatcher(self._target_pattern)
        else:
            self._matcher = matcher(self._target_pattern)
        self._verbose = verbose
        self.name = name

    def __str__(self) -> str:
        return self.name if self.name else "Anonymous Pattern"

    def match(
        self,
        model: ir.Model,
        graph_or_function: ir.Graph | ir.Function,
        node: ir.Node,
        *,
        verbose: int | None = None,
        check_nodes_are_removable: bool = True,
        tracer: _basics.MatchingTracer | None = None,
    ) -> _basics.MatchResult | None:
        """Check if the node matches the pattern and return the match result.

        Args:
            model: The model containing the graph or function.
            graph_or_function: The graph or function to match against.
            node: The node to try to match the pattern against.
            verbose: The verbosity level of messages.
            check_nodes_are_removable: If True, validate that matched nodes can be safely removed.
            tracer: The tracer for debugging.

        Returns:
            MatchResult if the pattern matches successfully and passes the condition function,
            None otherwise.
        """
        if verbose and verbose > 2:
            print(f"[match] {self}")
        verbose = verbose if verbose is not None else self._verbose
        match = self._matcher.match(
            model,
            graph_or_function,
            node,
            verbose=verbose,
            remove_nodes=check_nodes_are_removable,
        )
        if match:
            context = _basics.MatchContext(model, graph_or_function, node, match)
            for var in self._target_pattern.inputs:
                if var.name is not None:
                    if var.name not in match.bindings:
                        match.bind(var.name, None)

            # Perform value/node level checks before condition function
            def fail(check_result, default_message, failure_object=None):
                """Local utility to handle check failures consistently."""
                if isinstance(check_result, _basics.MatchResult):
                    match.fail(
                        check_result.reason,
                        check_result.failure_nodes_and_values,
                    )
                else:
                    match.fail(default_message, failure_object)
                if tracer:
                    tracer.log(
                        self,  # type: ignore[arg-type]
                        graph_or_function,
                        node,
                        match,
                        _basics.MatchStatus.CONDITION_FAILED,
                    )
                return None

            def wrap_try(f):
                """Encapsulates try-except pattern for check functions."""

                def wrapped(*args, **kwargs):
                    try:
                        return f(*args, **kwargs)
                    except _basics.MatchFailureError as e:
                        result = _basics.MatchResult()
                        result.fail(e.reason, list(e.failure_sources))
                        return result

                return wrapped

            # Check node-level checkers
            for pattern_node, ir_node in match.node_bindings.items():
                if pattern_node.check_method is not None:
                    check_result = wrap_try(pattern_node.check_method)(context, ir_node)
                    if not check_result:
                        return fail(
                            check_result,
                            f"Node-level check failed for pattern node {pattern_node}",
                            ir_node,
                        )

            # Check value-level checkers
            for pattern_value, ir_value in match.value_bindings.items():
                if pattern_value.check_method is not None:
                    check_result = wrap_try(pattern_value.check_method)(context, ir_value)
                    if not check_result:
                        return fail(
                            check_result,
                            f"Value-level check failed for pattern value {pattern_value}",
                            ir_value,
                        )

            check_match_result = wrap_try(self._condition_function)(context, **match.bindings)
            if not check_match_result:
                # If check function was provided, but it failed, return the reason for failure to the tracer.
                return fail(check_match_result, "Condition function check failed")
            if tracer:
                tracer.log(self, graph_or_function, node, match, _basics.MatchStatus.SUCCESS)  # type: ignore[arg-type]
            return match
        if tracer:
            tracer.log(self, graph_or_function, node, match, _basics.MatchStatus.NO_MATCH)  # type: ignore[arg-type]
        return match


class ReplacementPatternFunction:
    """The replacement pattern that will replace the targeted pattern.

    Attributes:
        function (Callable): The replacement function that will be used to replace the matched pattern.
    """

    def __init__(self, function) -> None:
        self._function = function

    def get_replacement(self, match: _basics.MatchResult) -> ReplacementSubgraph | None:
        context = RewriterContext()
        try:
            new_outputs = self._function(context, **match.bindings)
        except _basics.MatchFailureError as e:
            match.fail(e.reason, list(e.failure_sources))
            return None
        # Support the same failure conventions as check functions for uniformity:
        # - None or False: indicates failure without a reason (not recommended)
        # - Falsy MatchResult: failure with reason/source info (recommended)
        # - MatchFailureError exception: failure with reason/source info (recommended)
        if new_outputs is None or new_outputs is False:
            return None
        if isinstance(new_outputs, _basics.MatchResult):
            if not new_outputs:
                # A falsy MatchResult is the recommended way to signal failure with
                # reason/source information from a replacement function.
                match.fail(
                    new_outputs.reason,
                    new_outputs.failure_nodes_and_values,
                )
                return None
            # A truthy MatchResult should never be returned from a replacement
            # function. Treat this as a programmer error to avoid silent failures.
            raise TypeError(
                "Replacement function returned a truthy MatchResult. "
                "Replacement functions should either return None/False for a "
                "generic failure, return a *falsy* MatchResult to provide "
                "failure details, or raise MatchFailureError."
            )
        if not isinstance(new_outputs, Sequence):
            new_outputs = [new_outputs]
        return ReplacementSubgraph(
            match, new_outputs, context.nodes, context.initializers, context.used_opsets
        )


def _update_opset_imports(
    graph_or_function: ir.Graph | ir.Function, delta: ReplacementSubgraph
):
    imports = graph_or_function.opset_imports
    for domain, version in delta.used_opsets:
        if domain not in imports:
            # use 1 as default version if not explicitly specified
            imports[domain] = version if version is not None else 1
        elif version is not None and version != imports[domain]:
            raise ValueError(
                f"Multiple versions of opset {domain} used. "
                f"Expected version {imports[domain]}, but got {version}."
            )


class RewriteRule(Pattern):
    def __init__(
        self,
        target_pattern: _pattern_ir.GraphPattern | Callable,
        replacement_pattern: ReplacementPatternFunction | Callable,
        condition_function: Callable | None = None,
        matcher: _matcher.PatternMatcher
        | Callable[[_pattern_ir.GraphPattern], _matcher.PatternMatcher]
        | None = None,
        verbose: int = 0,
        name: str | None = None,
        remove_nodes: bool = True,
        graph_pre_visitor: Callable[[], None] | None = None,
        graph_post_visitor: Callable[[], None] | None = None,
        as_function: bool = False,
    ) -> None:
        """Create a rewrite rule.

        Args:
            target_pattern: The _pattern_ir.GraphPattern that will be matched against the IR.
                If a callable is provided, it will be converted to a _pattern_ir.GraphPattern.
            replacement_pattern: The ReplacementPatternFunction that will be used to
                replace the matched pattern. If a callable is provided, it will be
                converted to a ReplacementPatternFunction.
            condition_function: The condition function that will be used to check if
                the pattern match found should be rewritten.
            matcher: The pattern matcher that will be used to match the pattern.
                If not provided, a default matcher will be used.
            verbose: The verbosity level of the rule.
            name: An optional name for the pattern that will show up in verbose logging.
            remove_nodes: If True, the matched nodes will be removed from the graph.
            graph_pre_visitor: A function that will be called before applying the
                rewriting to the top-level graph or a function.
            graph_post_visitor: A function that will be called after the rewriting
                is complete for a graph or function.
            as_function: If True, the matched nodes will be extracted into a model
                local function. This is only supported when remove_nodes=True and
                when the replacement subgraph has a single node, representing the
                function call.
        """
        if as_function and not remove_nodes:
            raise ValueError("as_function=True is only supported when remove_nodes=True.")

        # Initialize the base pattern matching functionality
        super().__init__(target_pattern, condition_function, matcher, verbose, name)

        if not isinstance(replacement_pattern, ReplacementPatternFunction):
            replacement_pattern = ReplacementPatternFunction(replacement_pattern)
        self._replacement_pattern = replacement_pattern
        self.remove_nodes = remove_nodes
        self.graph_pre_visitor = graph_pre_visitor
        self.graph_post_visitor = graph_post_visitor
        self.as_function = as_function

    def __str__(self) -> str:
        return self.name if self.name else "Anonymous Rule"

    def try_rewrite(
        self,
        model: ir.Model,
        graph_or_function: ir.Graph | ir.Function,
        node: ir.Node,
        *,
        verbose: int | None = None,
        tracer: _basics.MatchingTracer | None = None,
    ) -> ReplacementSubgraph | None:
        """If the node matches the pattern, then replace the node with the replacement pattern."""
        # Use the inherited match method from Pattern
        match = self.match(
            model,
            graph_or_function,
            node,
            verbose=verbose,
            check_nodes_are_removable=self.remove_nodes,
            tracer=tracer,
        )
        if not match:
            return None

        replacement_subgraph = self._replacement_pattern.get_replacement(match)
        if replacement_subgraph is None:
            if tracer:
                tracer.log(
                    self,
                    graph_or_function,
                    node,
                    match,
                    _basics.MatchStatus.REPLACEMENT_FAILED,
                )
            return None
        if len(replacement_subgraph.new_outputs) != self._target_pattern.num_outputs:
            raise ValueError(
                f"Number of outputs from replacement function does not match the number of outputs from the target pattern. "
                f"Expected {self._target_pattern.num_outputs}, but got {len(replacement_subgraph.new_outputs)}."
            )
        # TODO(rama): Remove the opset imports from deleted nodes?
        _update_opset_imports(graph_or_function, replacement_subgraph)
        _update_opset_imports(model.graph, replacement_subgraph)
        return replacement_subgraph

    def apply_to_model(
        self,
        model: ir.Model,
        *,
        commute: bool = False,
        verbose: int | None = None,
        tracer: _basics.MatchingTracer | None = None,
    ):
        # A convenience method to apply the rule to a model. We use a RewriteRuleSet to
        # handle commutative rules.
        return RewriteRuleSet([self], commute=commute).apply_to_model(
            model, verbose=verbose, tracer=tracer
        )

    def commute(self) -> Sequence[RewriteRule]:
        def replace_pattern(new_pattern):
            """Return a shallow copy of self with node_pattern replaced by new_pattern."""
            # TODO(rama): Maybe we should use a better alternative to construct new matcher.
            matcher_class = type(self._matcher)
            return RewriteRule(
                new_pattern,
                self._replacement_pattern,
                self._condition_function,
                matcher_class(new_pattern),
                self._verbose,
                self.name,
                self.remove_nodes,
                self.graph_pre_visitor,
                self.graph_post_visitor,
                self.as_function,
            )

        return [replace_pattern(p) for p in self._target_pattern.commute()]


class PatternBase(abc.ABC):
    """Base class for implementing pattern matching as a class.

    This class encapsulates the pattern definition and condition checking
    without the replacement functionality.

    Example::

        class TransposePattern(PatternBase):
            def pattern(cls, op, x, perm):
                return op.Transpose(x, perm=perm)

            def check(cls, context, x: ir.Value, perm: ir.Attr) -> bool:
                if perm.is_ref():
                    return False
                if perm.type == ir.AttributeType.INTS:
                    if list(perm.as_ints()) == list(range(len(perm.as_ints()))):
                        return True
                return False
    """

    def __init__(self, name: str | None = None, **kwargs) -> None:
        self.name = name or self.__class__.__name__
        # Initialize to None and create on demand to avoid construction order issues
        self._compiled_pattern: Pattern | None = None
        self._pattern_kwargs = kwargs

    @abc.abstractmethod
    def pattern(self, op, *args, **kwargs):
        raise NotImplementedError("Method 'pattern' must be implemented by derived class.")

    def check(self, op, *args, **kwargs) -> _basics.MatchResult:
        """Default check function that returns a _basics.MatchResult object with success always set to True."""
        return _basics.MatchResult()

    def match(
        self,
        model: ir.Model,
        graph_or_function: ir.Graph | ir.Function,
        node: ir.Node,
        *,
        verbose: int | None = None,
        check_nodes_are_removable: bool = True,
        tracer: _basics.MatchingTracer | None = None,
    ) -> _basics.MatchResult | None:
        """Check if the node matches the pattern and return the match result.

        Args:
            model: The model containing the graph or function.
            graph_or_function: The graph or function to match against.
            node: The node to try to match the pattern against.
            verbose: The verbosity level of messages.
            check_nodes_are_removable: If True, validate that matched nodes can be safely removed.
            tracer: The tracer for debugging.

        Returns:
            MatchResult if the pattern matches successfully and passes the condition function,
            None otherwise.
        """
        # Create the compiled pattern on demand if not already created
        if self._compiled_pattern is None:
            self._compiled_pattern = Pattern(
                self.pattern, self.check, name=self.name, **self._pattern_kwargs
            )
        return self._compiled_pattern.match(
            model,
            graph_or_function,
            node,
            verbose=verbose,
            check_nodes_are_removable=check_nodes_are_removable,
            tracer=tracer,
        )


class RewriteRuleClassBase(PatternBase):
    """Base class for implementing rewrite rules as a class.

    Example::

        class TransposeIdentity(RewriteRuleClassBase):
            def pattern(cls, op, x, perm):
                return op.Transpose(x, perm=perm)

            def check(cls, context, x: ir.Value, perm: ir.Attr) -> bool:
                if perm.is_ref():
                    return False
                if perm.type == ir.AttributeType.INTS:
                    if list(perm.as_ints()) == list(range(len(perm.as_ints()))):
                        return True
                return False

            def rewrite(cls, op, x: ir.Value, perm: ir.Attr | None = None):
                return op.Identity(x)

        # Then use
        # TransposeIdentity.rule()
        # to create a RewriteRule object.

    """

    @classmethod
    def rule(cls, *args, **kwargs):
        instance = cls(*args, **kwargs)
        return RewriteRule(
            instance.pattern,
            instance.rewrite,
            instance.check,
            name=instance.name,
            remove_nodes=instance.remove_nodes,
            graph_pre_visitor=instance.setup,
            graph_post_visitor=instance.cleanup,
            as_function=instance.as_function,
        )

    def __init__(
        self, name: str | None = None, remove_nodes: bool = True, as_function: bool = False
    ) -> None:
        super().__init__(name)
        self.remove_nodes = remove_nodes
        self.as_function = as_function

    @abc.abstractmethod
    def rewrite(self, op, *args, **kwargs):
        raise NotImplementedError("Method 'rewrite' must be implemented by derived class.")

    def setup(self):
        """Optional setup function that can be overridden by derived classes.

        Used to do per model/function initialization.
        """
        return

    def cleanup(self):
        """Optional cleanup function that can be overridden by derived classes.

        Used to do per model/function cleanup.
        """
        return


def _copy_for_function(
    inputs: Sequence[ir.Value | None], nodes: Sequence[ir.Node], outputs: Sequence[ir.Value]
):
    """Utility function to extract a subgraph out as a function."""
    value_map: dict[ir.Value, ir.Value] = {}
    function_inputs: list[ir.Value] = []
    constant_nodes: list[ir.Node] = []
    for input in inputs:
        # Create a function input (formal-parameter value) to represent this value:
        new_value = (
            ir.Value(
                name=input.name,
                shape=input.shape,
                type=input.type,
                doc_string=input.doc_string,
            )
            if input
            else ir.Value()  # dummy parameter for a None input
        )
        if input is not None:
            value_map[input] = new_value
        function_inputs.append(new_value)

    def copy_value(value: ir.Value | None) -> ir.Value | None:
        if value is None:
            return None
        if value not in value_map:
            const_value = value.const_value
            if const_value is not None:
                # create a Constant node to represent the value
                value_attr = ir.AttrTensor("value", const_value)
                const_node = ir.Node("", "Constant", [], [value_attr])
                constant_nodes.append(const_node)
                value_map[value] = result = const_node.outputs[0]
                return result
            raise ValueError(f"Value {value} not found in value_map.")
        return value_map[value]

    def copy_attr_value(attr: ir.Attr) -> ir.Attr:
        if attr.is_ref():
            # No need to support this currently, as rewriting inside a function is
            # not used, as it has several challenges.
            raise NotImplementedError("RefAttr not supported.")
        if attr.type in {ir.AttributeType.GRAPH, ir.AttributeType.GRAPHS}:
            # No need to support this currently, as rewriting control-flow constructs
            # is not used and has several challenges.
            raise NotImplementedError("Graph attributes not supported.")
        # Primitive attributes are immutable by design and can be shared.
        return attr

    def copy_node(node: ir.Node) -> ir.Node:
        new_inputs = [copy_value(v) for v in node.inputs]
        new_attributes = [copy_attr_value(v) for v in node.attributes.values()]
        new_node = ir.Node(
            node.domain,
            node.op_type,
            new_inputs,
            new_attributes,
            overload=node.overload,
            num_outputs=len(node.outputs),
            graph=None,
            name=node.name,
            doc_string=node.doc_string,  # type: ignore
            metadata_props=node.metadata_props.copy(),
        )
        new_outputs = new_node.outputs
        for i, output in enumerate(node.outputs):
            value_map[output] = new_outputs[i]
            if output.name is not None:
                new_outputs[i].name = output.name
        return new_node

    function_nodes = [copy_node(node) for node in nodes]
    function_outputs = [copy_value(v) for v in outputs]
    return (function_inputs, constant_nodes + function_nodes, function_outputs)


def _get_new_overload(model: ir.Model, domain: str, name: str) -> str:
    """Get a new overload for the given domain and name.

    Args:
        model: The model to which the new overload will be added.
        domain: The domain of the new overload.
        name: The opname of the new overload.

    Returns:
        The new overload name.
    """
    existing_functions = model.functions
    # Just a simple implementation for now
    overload = 1
    while True:
        overload_name = str(overload)
        if (domain, name, overload_name) not in existing_functions:
            return overload_name
        overload += 1


_default_metadata_merger: metadata_merger.MetadataMerger = metadata_merger.MetadataMerger(
    {RULE_NAME_TAG: metadata_merger.comma_separator_merger}
)

# TODO(rama): Generalize this to support custom metadata mergers. For now, we just allow
# enabling/disabling the default merger.
merge_metadata: bool = True


class RewriteRuleSet:
    def __init__(self, rules: Sequence[RewriteRule], *, commute: bool = False) -> None:
        if not rules:
            raise ValueError("rules must contain at least one rule")
        if commute:
            rules = list(itertools.chain.from_iterable([rule.commute() for rule in rules]))
        self.rules = rules
        # We call remove_unused_nodes at end of rewriting if there is any rule that does
        # NOT remove nodes (immediately when it is applied)
        self.remove_unused_nodes = any(not rule.remove_nodes for rule in rules)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.rules})"

    def _apply_to_graph_or_function(
        self,
        model: ir.Model,
        graph_or_function: ir.Graph | ir.Function,
        *,
        verbose: int | None,
        tracer: _basics.MatchingTracer | None = None,
    ) -> int:
        """
        Apply the rewrite rules to the given graph or function.

        Args:
            model: The model to which the rewrite rules are applied.
            graph_or_function: The graph or function to which the rewrite rules are applied.
            verbose: The verbosity level. Defaults to None.
            tracer: The tracer for debugging. Defaults to None.

        Returns:
            The number of rewrite rules applied.
        """
        count = 0

        for rule in self.rules:
            if rule.graph_pre_visitor:
                rule.graph_pre_visitor()

        for node in graph_or_function:
            for rule in self.rules:
                delta = rule.try_rewrite(
                    model, graph_or_function, node, verbose=verbose, tracer=tracer
                )
                if delta is None or tracer is not None:
                    continue
                assert isinstance(delta, ReplacementSubgraph)
                if delta.new_initializers:
                    if isinstance(graph_or_function, ir.Function):
                        # TODO(rama): Can't add initializers to functions. But currently this is not
                        # an issue, as we apply inlining before applying rewrite rules.
                        if verbose:
                            print(
                                f"Rewrites adding initializers not supported for functions: {rule}"
                            )
                        continue
                    initializers = graph_or_function.initializers
                    for initializer in delta.new_initializers:
                        if initializer.name in initializers:
                            if verbose:
                                print(f"Initializer {initializer.name} already exists.")
                            continue
                    for initializer in delta.new_initializers:
                        initializers[initializer.name] = initializer  # type: ignore[index]
                # TODO: This does not yet handle the problem of determining the correct insertion point
                # for inserted nodes in the case of patterns with multiple output-nodes. The following
                # is sufficient for patterns with a single output-node "node", which can serve as the
                # insertion-point.
                onnxscript.optimizer.basic_constant_propagation(delta.new_nodes)
                if rule.as_function:
                    # Create a function out of a copy of the matched nodes
                    if len(delta.new_nodes) != 1:
                        raise ValueError(
                            "as_function=True is only supported for patterns with a single replacement node."
                        )
                    call_node = delta.new_nodes[0]
                    domain = call_node.domain
                    name = call_node.op_type
                    overload = _get_new_overload(model, domain, name)
                    call_node.overload = overload

                    # Create topologically sorted list of nodes to be replaced.
                    unsorted_nodes = set(delta.match.nodes)
                    original_nodes = [n for n in graph_or_function if n in unsorted_nodes]
                    # Create new inputs/nodes/outputs for the function
                    inputs, nodes, outputs = _copy_for_function(
                        call_node.inputs, original_nodes, delta.match.outputs
                    )

                    used_domains: set[str] = {node.domain for node in original_nodes}
                    parent_opset_imports = graph_or_function.opset_imports
                    used_opset_imports = {
                        k: v for k, v in parent_opset_imports.items() if k in used_domains
                    }

                    graph = ir.Graph(
                        inputs, outputs, nodes=nodes, opset_imports=used_opset_imports
                    )
                    f = ir.Function(domain, name, overload, graph=graph, attributes=())
                    model.functions[f.identifier()] = f

                if verbose:
                    name = f"{rule.name}: " if rule.name else ""
                    print(f"----{name}Matched Nodes----")
                    _ir_utils.display_nodes(delta.match.nodes)
                    print("++++Replacement Nodes++++")
                    _ir_utils.display_nodes(delta.new_nodes)
                    print("++++End Replacement Nodes++++")

                # Capture rewrite rule name as metadata.
                # TODO(rama): This is just a basic version. We may wish to compose "source" metadata
                # from multiple rules in future.
                if rule.name:
                    for n in delta.new_nodes:
                        n.metadata_props[RULE_NAME_TAG] = rule.name

                convenience.replace_nodes_and_values(
                    graph_or_function,
                    node,
                    delta.match.nodes if rule.remove_nodes else [],
                    delta.new_nodes,
                    delta.match.outputs,
                    delta.new_outputs,
                )

                if merge_metadata:
                    _default_metadata_merger.copy_merged_metadata(
                        delta.match.nodes, delta.new_nodes
                    )

                count += 1
                break

            # Apply rewrite rules to subgraphs of the node.
            for attr in node.attributes.values():
                if attr.type == ir.AttributeType.GRAPH:
                    count += self._apply_to_graph_or_function(
                        model, attr.value, verbose=verbose, tracer=tracer
                    )
                elif attr.type == ir.AttributeType.GRAPHS:
                    for graph in attr.value:
                        count += self._apply_to_graph_or_function(
                            model, graph, verbose=verbose, tracer=tracer
                        )

        for rule in self.rules:
            if rule.graph_post_visitor:
                rule.graph_post_visitor()

        return count

    def apply_to_model(
        self,
        model: ir.Model,
        *,
        verbose: int | None = None,
        tracer: _basics.MatchingTracer | None = None,
    ) -> int:
        """Apply the rewrite rules in the set to the model.

        Args:
            model: The model to which the rewrite rules are applied.
            verbose: The verbosity level of messages. Defaults to None.
            tracer: if specified, no changes are made to the model, only
                information about the best matches found is computed.

        Returns:
            The number of applications of rewrite rules.
        """
        assert isinstance(model, ir.Model)
        onnxscript.optimizer.basic_constant_propagation(model.graph)
        # Rewriting may introduce new functions. In the following loop,
        # we restrict rewriting to original functions, not newly introduced ones.
        original_functions = list(model.functions.values())
        count = self._apply_to_graph_or_function(
            model, model.graph, verbose=verbose, tracer=tracer
        )
        for function in original_functions:
            onnxscript.optimizer.basic_constant_propagation(function)
            count += self._apply_to_graph_or_function(
                model, function, verbose=verbose, tracer=tracer
            )
        if self.remove_unused_nodes:
            onnxscript.optimizer.remove_unused_nodes(model)
        return count

    def __iter__(self):
        yield from self.rules
